# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import os
import textwrap
from unittest import mock

from testtools.matchers import Contains, Equals, HasLength

import snapcraft
from snapcraft.internal import lifecycle, pluginhandler, states, steps
from snapcraft.internal.lifecycle._status_cache import StatusCache

from . import LifecycleTestBase

# TODO refactor this entire module, the current test time is around 15 minutes.


def pull_order(part_name=None):
    events = []

    if not part_name or part_name in ("main", "dependent", "nested-dependent"):
        events.extend([dict(part="main", step=steps.PULL)])

    if not part_name or part_name in ("dependent", "nested-dependent"):
        events.extend(
            [
                dict(part="main", step=steps.BUILD),
                dict(part="main", step=steps.STAGE),
                dict(part="dependent", step=steps.PULL),
            ]
        )

    if not part_name or part_name == "nested-dependent":
        events.extend(
            [
                dict(part="dependent", step=steps.BUILD),
                dict(part="dependent", step=steps.STAGE),
                dict(part="nested-dependent", step=steps.PULL),
            ]
        )

    return events


def build_order(part_name=None):
    events = []

    if not part_name or part_name in ("main", "dependent", "nested-dependent"):
        events.extend(
            [dict(part="main", step=steps.PULL), dict(part="main", step=steps.BUILD)]
        )

    if not part_name or part_name in ("dependent", "nested-dependent"):
        events.extend(
            [
                dict(part="main", step=steps.STAGE),
                dict(part="dependent", step=steps.PULL),
                dict(part="dependent", step=steps.BUILD),
            ]
        )

    if not part_name or part_name == "nested-dependent":
        events.extend(
            [
                dict(part="dependent", step=steps.STAGE),
                dict(part="nested-dependent", step=steps.PULL),
                dict(part="nested-dependent", step=steps.BUILD),
            ]
        )

    return events


def stage_order(part_name=None):
    events = []

    if not part_name or part_name in ("main", "dependent", "nested-dependent"):
        events.extend(
            [
                dict(part="main", step=steps.PULL),
                dict(part="main", step=steps.BUILD),
                dict(part="main", step=steps.STAGE),
            ]
        )

    if not part_name or part_name in ("dependent", "nested-dependent"):
        events.extend(
            [
                dict(part="dependent", step=steps.PULL),
                dict(part="dependent", step=steps.BUILD),
                dict(part="dependent", step=steps.STAGE),
            ]
        )

    if not part_name or part_name == "nested-dependent":
        events.extend(
            [
                dict(part="nested-dependent", step=steps.PULL),
                dict(part="nested-dependent", step=steps.BUILD),
                dict(part="nested-dependent", step=steps.STAGE),
            ]
        )

    return events


def prime_order(part_name=None):
    events = stage_order(part_name)

    if not part_name or part_name in ("main", "dependent", "nested-dependent"):
        events.append(dict(part="main", step=steps.PRIME))

    if not part_name or part_name in ("dependent", "nested-dependent"):
        events.append(dict(part="dependent", step=steps.PRIME))

    if not part_name or part_name == "nested-dependent":
        events.append(dict(part="nested-dependent", step=steps.PRIME))

    return events


class OrderTestBase(LifecycleTestBase):
    def setUp(self):
        super().setUp()

        self.project_config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                parts:
                  main:
                    plugin: nil
                  dependent:
                    plugin: nil
                    after: [main]
                  nested-dependent:
                    plugin: nil
                    after: [dependent]
                """
            )
        )

        # Set the option to automatically clean dirty/outdated steps
        with snapcraft.config.CLIConfig() as cli_config:
            cli_config.set_outdated_step_action(
                snapcraft.config.OutdatedStepAction.CLEAN
            )

    def set_attributes(self, kwargs):
        self.__dict__.update(kwargs)

    def get_run_order(self):
        # Let's determine run order by using the timestamp of state files
        actual_order = []
        for part_name in ("main", "dependent", "nested-dependent"):
            state_dir = os.path.join(self.parts_dir, part_name, "state")
            with contextlib.suppress(FileNotFoundError):
                for step_name in os.listdir(state_dir):
                    path = os.path.join(state_dir, step_name)
                    actual_order.append(
                        {
                            "part": part_name,
                            "step": steps.get_step_by_name(step_name),
                            "timestamp": os.stat(path).st_mtime,
                        }
                    )

        # Now sort by timestamp to get real order
        return sorted(actual_order, key=lambda k: k["timestamp"])

    def assert_parts_dirty(self, expected_parts, hint):
        cache = StatusCache(self.project_config)
        dirty_parts = []
        for part in self.project_config.parts.all_parts:
            for step in steps.STEPS:
                if cache.get_dirty_report(part, step):
                    dirty_parts.append(dict(part=part.name, step=step))
        self.assertThat(dirty_parts, HasLength(len(expected_parts)), hint)
        for expected in expected_parts:
            self.assertThat(dirty_parts, Contains(expected), hint)

    def assert_parts_cleaned(self, earlier_parts, current_parts, expected_parts, hint):
        cleaned_parts = []
        for earlier in earlier_parts:
            earlier_part = earlier["part"]
            earlier_step = earlier["step"]
            found = False
            for current in current_parts:
                if earlier_part == current["part"] and earlier_step == current["step"]:
                    found = True
                    break
            if not found:
                cleaned_parts.append(dict(part=earlier_part, step=earlier_step))

        self.assertThat(cleaned_parts, HasLength(len(expected_parts)), hint)
        for expected in expected_parts:
            self.assertThat(cleaned_parts, Contains(expected), hint)

    def assert_run_order_equal(self, actual_order, expected_order, hint):
        self.assertThat(actual_order, HasLength(len(expected_order)), hint)
        for actual, expected in zip(actual_order, expected_order):
            actual_tuple = (actual["part"], actual["step"])
            expected_tuple = (expected["part"], expected["step"])
            self.assertThat(
                actual_tuple,
                Equals(expected_tuple),
                "expected {} {}".format(expected_tuple, hint),
            )


class ReStepOrderTestBase(OrderTestBase):
    def run_test(self):
        lifecycle.execute(
            self.initial_step, self.project_config, part_names=self.initial_parts
        )

        initial_order = self.get_run_order()
        self.assert_run_order_equal(initial_order, self.expected_initial, "(initial)")

        lifecycle.execute(
            self.test_step, self.project_config, part_names=self.test_parts
        )

        current_order = self.get_run_order()
        test_order = [o for o in current_order if o not in initial_order]
        self.assert_run_order_equal(test_order, self.expected_test, "(test)")

        self.assert_parts_cleaned(
            initial_order, current_order, self.expected_test_cleaned, "(cleaned)"
        )

        self.assert_run_order_equal(
            self.get_run_order(), self.expected_final, "(final)"
        )

        self.assert_parts_dirty(self.expected_dirty, "(dirty)")


class MainTest(ReStepOrderTestBase):
    def test_pull_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.PULL,
                "initial_parts": ["main"],
                "test_parts": ["main"],
                "test_step": steps.PULL,
                "expected_initial": pull_order("main"),
                "expected_test": pull_order("main"),
                "expected_test_cleaned": [],
                "expected_final": pull_order("main"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_build_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.BUILD,
                "initial_parts": ["main"],
                "test_parts": ["main"],
                "test_step": steps.PULL,
                "expected_initial": build_order("main"),
                "expected_test": pull_order("main"),
                "expected_test_cleaned": [dict(part="main", step=steps.BUILD)],
                "expected_final": pull_order("main"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_stage_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.STAGE,
                "initial_parts": ["main"],
                "test_parts": ["main"],
                "test_step": steps.PULL,
                "expected_initial": stage_order("main"),
                "expected_test": pull_order("main"),
                "expected_test_cleaned": [
                    dict(part="main", step=steps.BUILD),
                    dict(part="main", step=steps.STAGE),
                ],
                "expected_final": pull_order("main"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_prime_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": ["main"],
                "test_parts": ["main"],
                "test_step": steps.PULL,
                "expected_initial": prime_order("main"),
                "expected_test": pull_order("main"),
                "expected_test_cleaned": [
                    dict(part="main", step=steps.BUILD),
                    dict(part="main", step=steps.STAGE),
                    dict(part="main", step=steps.PRIME),
                ],
                "expected_final": pull_order("main"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_all_pull_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.PULL,
                "initial_parts": [],
                "test_parts": ["main"],
                "test_step": steps.PULL,
                "expected_initial": pull_order(),
                "expected_test": pull_order("main"),
                "expected_test_cleaned": [
                    dict(part="main", step=steps.BUILD),
                    dict(part="main", step=steps.STAGE),
                ],
                "expected_final": [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                ]
                + pull_order("main"),
                "expected_dirty": [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                ],
            }
        )

        self.run_test()

    def test_all_build_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.BUILD,
                "initial_parts": [],
                "test_parts": ["main"],
                "test_step": steps.PULL,
                "expected_initial": build_order(),
                "expected_test": pull_order("main"),
                "expected_test_cleaned": [
                    dict(part="main", step=steps.BUILD),
                    dict(part="main", step=steps.STAGE),
                ],
                "expected_final": [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                ]
                + pull_order("main"),
                "expected_dirty": [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                ],
            }
        )

        self.run_test()

    def test_all_stage_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.STAGE,
                "initial_parts": [],
                "test_parts": ["main"],
                "test_step": steps.PULL,
                "expected_initial": stage_order(),
                "expected_test": pull_order("main"),
                "expected_test_cleaned": [
                    dict(part="main", step=steps.BUILD),
                    dict(part="main", step=steps.STAGE),
                ],
                "expected_final": [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                ]
                + pull_order("main"),
                "expected_dirty": [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                ],
            }
        )

        self.run_test()

    def test_all_prime_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": [],
                "test_parts": ["main"],
                "test_step": steps.PULL,
                "expected_initial": prime_order(),
                "expected_test": pull_order("main"),
                "expected_test_cleaned": [
                    dict(part="main", step=steps.BUILD),
                    dict(part="main", step=steps.STAGE),
                    dict(part="main", step=steps.PRIME),
                ],
                "expected_final": [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ]
                + pull_order("main"),
                "expected_dirty": [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )

        self.run_test()

    def test_build_rebuild(self):
        self.set_attributes(
            {
                "initial_step": steps.BUILD,
                "initial_parts": ["main"],
                "test_parts": ["main"],
                "test_step": steps.BUILD,
                "expected_initial": build_order("main"),
                "expected_test": [dict(part="main", step=steps.BUILD)],
                "expected_test_cleaned": [],
                "expected_final": build_order("main"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_stage_rebuild(self):
        self.set_attributes(
            {
                "initial_step": steps.STAGE,
                "initial_parts": ["main"],
                "test_parts": ["main"],
                "test_step": steps.BUILD,
                "expected_initial": stage_order("main"),
                "expected_test": [dict(part="main", step=steps.BUILD)],
                "expected_test_cleaned": [dict(part="main", step=steps.STAGE)],
                "expected_final": build_order("main"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_prime_rebuild(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": ["main"],
                "test_parts": ["main"],
                "test_step": steps.BUILD,
                "expected_initial": prime_order("main"),
                "expected_test": [dict(part="main", step=steps.BUILD)],
                "expected_test_cleaned": [
                    dict(part="main", step=steps.STAGE),
                    dict(part="main", step=steps.PRIME),
                ],
                "expected_final": build_order("main"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_all_build_rebuild(self):
        self.set_attributes(
            {
                "initial_step": steps.BUILD,
                "initial_parts": [],
                "test_parts": ["main"],
                "test_step": steps.BUILD,
                "expected_initial": build_order(),
                "expected_test": [dict(part="main", step=steps.BUILD)],
                "expected_test_cleaned": [dict(part="main", step=steps.STAGE)],
                "expected_final": [
                    dict(part="main", step=steps.PULL),
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="main", step=steps.BUILD),
                ],
                "expected_dirty": [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                ],
            }
        )

        self.run_test()

    def test_all_stage_rebuild(self):
        self.set_attributes(
            {
                "initial_step": steps.STAGE,
                "initial_parts": [],
                "test_parts": ["main"],
                "test_step": steps.BUILD,
                "expected_initial": stage_order(),
                "expected_test": [dict(part="main", step=steps.BUILD)],
                "expected_test_cleaned": [dict(part="main", step=steps.STAGE)],
                "expected_final": [
                    dict(part="main", step=steps.PULL),
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="main", step=steps.BUILD),
                ],
                "expected_dirty": [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                ],
            }
        )

        self.run_test()

    def test_all_prime_rebuild(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": [],
                "test_parts": ["main"],
                "test_step": steps.BUILD,
                "expected_initial": prime_order(),
                "expected_test": [dict(part="main", step=steps.BUILD)],
                "expected_test_cleaned": [
                    dict(part="main", step=steps.STAGE),
                    dict(part="main", step=steps.PRIME),
                ],
                "expected_final": [
                    dict(part="main", step=steps.PULL),
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                    dict(part="main", step=steps.BUILD),
                ],
                "expected_dirty": [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )

        self.run_test()

    def test_stage_restage(self):
        self.set_attributes(
            {
                "initial_step": steps.STAGE,
                "initial_parts": ["main"],
                "test_parts": ["main"],
                "test_step": steps.STAGE,
                "expected_initial": stage_order("main"),
                "expected_test": [dict(part="main", step=steps.STAGE)],
                "expected_test_cleaned": [],
                "expected_final": stage_order("main"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_prime_restage(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": ["main"],
                "test_parts": ["main"],
                "test_step": steps.STAGE,
                "expected_initial": prime_order("main"),
                "expected_test": [dict(part="main", step=steps.STAGE)],
                "expected_test_cleaned": [dict(part="main", step=steps.PRIME)],
                "expected_final": stage_order("main"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_all_stage_restage(self):
        self.set_attributes(
            {
                "initial_step": steps.STAGE,
                "initial_parts": [],
                "test_parts": ["main"],
                "test_step": steps.STAGE,
                "expected_initial": stage_order(),
                "expected_test": [dict(part="main", step=steps.STAGE)],
                "expected_test_cleaned": [],
                "expected_final": build_order("main")
                + [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="main", step=steps.STAGE),
                ],
                "expected_dirty": [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                ],
            }
        )

        self.run_test()

    def test_all_prime_restage(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": [],
                "test_parts": ["main"],
                "test_step": steps.STAGE,
                "expected_initial": prime_order(),
                "expected_test": [dict(part="main", step=steps.STAGE)],
                "expected_test_cleaned": [dict(part="main", step=steps.PRIME)],
                "expected_final": build_order("main")
                + [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                    dict(part="main", step=steps.STAGE),
                ],
                "expected_dirty": [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )

        self.run_test()

    def test_prime_reprime(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": ["main"],
                "test_parts": ["main"],
                "test_step": steps.PRIME,
                "expected_initial": prime_order("main"),
                "expected_test": [dict(part="main", step=steps.PRIME)],
                "expected_test_cleaned": [],
                "expected_final": prime_order("main"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_all_prime_reprime(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": [],
                "test_parts": ["main"],
                "test_step": steps.PRIME,
                "expected_initial": prime_order(),
                "expected_test": [dict(part="main", step=steps.PRIME)],
                "expected_test_cleaned": [],
                "expected_final": stage_order()
                + [
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                    dict(part="main", step=steps.PRIME),
                ],
                "expected_dirty": [
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )

        self.run_test()


class DependentTest(ReStepOrderTestBase):
    def test_pull_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.PULL,
                "initial_parts": ["dependent"],
                "test_parts": ["dependent"],
                "test_step": steps.PULL,
                "expected_initial": pull_order("dependent"),
                "expected_test": [dict(part="dependent", step=steps.PULL)],
                "expected_test_cleaned": [],
                "expected_final": pull_order("dependent"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_build_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.BUILD,
                "initial_parts": ["dependent"],
                "test_parts": ["dependent"],
                "test_step": steps.PULL,
                "expected_initial": build_order("dependent"),
                "expected_test": [dict(part="dependent", step=steps.PULL)],
                "expected_test_cleaned": [dict(part="dependent", step=steps.BUILD)],
                "expected_final": pull_order("dependent"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_stage_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.STAGE,
                "initial_parts": ["dependent"],
                "test_parts": ["dependent"],
                "test_step": steps.PULL,
                "expected_initial": stage_order("dependent"),
                "expected_test": [dict(part="dependent", step=steps.PULL)],
                "expected_test_cleaned": [
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                ],
                "expected_final": pull_order("dependent"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_prime_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": ["dependent"],
                "test_parts": ["dependent"],
                "test_step": steps.PULL,
                "expected_initial": prime_order("dependent"),
                "expected_test": [dict(part="dependent", step=steps.PULL)],
                "expected_test_cleaned": [
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                ],
                "expected_final": prime_order("main")
                + [dict(part="dependent", step=steps.PULL)],
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_all_pull_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.PULL,
                "initial_parts": [],
                "test_parts": ["dependent"],
                "test_step": steps.PULL,
                "expected_initial": pull_order(),
                "expected_test": [dict(part="dependent", step=steps.PULL)],
                "expected_test_cleaned": [
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                ],
                "expected_final": stage_order("main")
                + [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.PULL),
                ],
                "expected_dirty": [dict(part="nested-dependent", step=steps.PULL)],
            }
        )

        self.run_test()

    def test_all_build_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.BUILD,
                "initial_parts": [],
                "test_parts": ["dependent"],
                "test_step": steps.PULL,
                "expected_initial": build_order(),
                "expected_test": [dict(part="dependent", step=steps.PULL)],
                "expected_test_cleaned": [
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                ],
                "expected_final": stage_order("main")
                + [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.PULL),
                ],
                "expected_dirty": [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                ],
            }
        )

        self.run_test()

    def test_all_stage_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.STAGE,
                "initial_parts": [],
                "test_parts": ["dependent"],
                "test_step": steps.PULL,
                "expected_initial": stage_order(),
                "expected_test": [dict(part="dependent", step=steps.PULL)],
                "expected_test_cleaned": [
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                ],
                "expected_final": stage_order("main")
                + [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PULL),
                ],
                "expected_dirty": [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                ],
            }
        )

        self.run_test()

    def test_all_prime_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": [],
                "test_parts": ["dependent"],
                "test_step": steps.PULL,
                "expected_initial": prime_order(),
                "expected_test": [dict(part="dependent", step=steps.PULL)],
                "expected_test_cleaned": [
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                ],
                "expected_final": stage_order("main")
                + [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="main", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                    dict(part="dependent", step=steps.PULL),
                ],
                "expected_dirty": [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )

        self.run_test()

    def test_build_rebuild(self):
        self.set_attributes(
            {
                "initial_step": steps.BUILD,
                "initial_parts": ["dependent"],
                "test_parts": ["dependent"],
                "test_step": steps.BUILD,
                "expected_initial": build_order("dependent"),
                "expected_test": [dict(part="dependent", step=steps.BUILD)],
                "expected_test_cleaned": [],
                "expected_final": build_order("dependent"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_stage_rebuild(self):
        self.set_attributes(
            {
                "initial_step": steps.STAGE,
                "initial_parts": ["dependent"],
                "test_parts": ["dependent"],
                "test_step": steps.BUILD,
                "expected_initial": stage_order("dependent"),
                "expected_test": [dict(part="dependent", step=steps.BUILD)],
                "expected_test_cleaned": [dict(part="dependent", step=steps.STAGE)],
                "expected_final": build_order("dependent"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_prime_rebuild(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": ["dependent"],
                "test_parts": ["dependent"],
                "test_step": steps.BUILD,
                "expected_initial": prime_order("dependent"),
                "expected_test": [dict(part="dependent", step=steps.BUILD)],
                "expected_test_cleaned": [
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                ],
                "expected_final": pull_order("dependent")
                + [
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.BUILD),
                ],
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_all_build_rebuild(self):

        self.set_attributes(
            {
                "initial_step": steps.BUILD,
                "initial_parts": [],
                "test_parts": ["dependent"],
                "test_step": steps.BUILD,
                "expected_initial": build_order(),
                "expected_test": [dict(part="dependent", step=steps.BUILD)],
                "expected_test_cleaned": [dict(part="dependent", step=steps.STAGE)],
                "expected_final": stage_order("main")
                + [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.BUILD),
                ],
                "expected_dirty": [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                ],
            }
        )

        self.run_test()

    def test_all_stage_rebuild(self):
        self.set_attributes(
            {
                "initial_step": steps.STAGE,
                "initial_parts": [],
                "test_parts": ["dependent"],
                "test_step": steps.BUILD,
                "expected_initial": stage_order(),
                "expected_test": [dict(part="dependent", step=steps.BUILD)],
                "expected_test_cleaned": [dict(part="dependent", step=steps.STAGE)],
                "expected_final": stage_order("main")
                + [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.BUILD),
                ],
                "expected_dirty": [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                ],
            }
        )

        self.run_test()

    def test_all_prime_rebuild(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": [],
                "test_parts": ["dependent"],
                "test_step": steps.BUILD,
                "expected_initial": prime_order(),
                "expected_test": [dict(part="dependent", step=steps.BUILD)],
                "expected_test_cleaned": [
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                ],
                "expected_final": pull_order("dependent")
                + [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="main", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                    dict(part="dependent", step=steps.BUILD),
                ],
                "expected_dirty": [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )

        self.run_test()

    def test_stage_restage(self):
        self.set_attributes(
            {
                "initial_step": steps.STAGE,
                "initial_parts": ["dependent"],
                "test_parts": ["dependent"],
                "test_step": steps.STAGE,
                "expected_initial": stage_order("dependent"),
                "expected_test": [dict(part="dependent", step=steps.STAGE)],
                "expected_test_cleaned": [],
                "expected_final": stage_order("dependent"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_prime_restage(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": ["dependent"],
                "test_parts": ["dependent"],
                "test_step": steps.STAGE,
                "expected_initial": prime_order("dependent"),
                "expected_test": [dict(part="dependent", step=steps.STAGE)],
                "expected_test_cleaned": [dict(part="dependent", step=steps.PRIME)],
                "expected_final": build_order("dependent")
                + [
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.STAGE),
                ],
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_all_stage_restage(self):
        self.set_attributes(
            {
                "initial_step": steps.STAGE,
                "initial_parts": [],
                "test_parts": ["dependent"],
                "test_step": steps.STAGE,
                "expected_initial": stage_order(),
                "expected_test": [dict(part="dependent", step=steps.STAGE)],
                "expected_test_cleaned": [],
                "expected_final": build_order("dependent")
                + [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.STAGE),
                ],
                "expected_dirty": [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                ],
            }
        )

        self.run_test()

    def test_all_prime_restage(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": [],
                "test_parts": ["dependent"],
                "test_step": steps.STAGE,
                "expected_initial": prime_order(),
                "expected_test": [dict(part="dependent", step=steps.STAGE)],
                "expected_test_cleaned": [dict(part="dependent", step=steps.PRIME)],
                "expected_final": build_order("dependent")
                + [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="main", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                    dict(part="dependent", step=steps.STAGE),
                ],
                "expected_dirty": [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )

        self.run_test()

    def test_prime_reprime(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": ["dependent"],
                "test_parts": ["dependent"],
                "test_step": steps.PRIME,
                "expected_initial": prime_order("dependent"),
                "expected_test": [dict(part="dependent", step=steps.PRIME)],
                "expected_test_cleaned": [],
                "expected_final": prime_order("dependent"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_all_prime_reprime(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": [],
                "test_parts": ["dependent"],
                "test_step": steps.PRIME,
                "expected_initial": prime_order(),
                "expected_test": [dict(part="dependent", step=steps.PRIME)],
                "expected_test_cleaned": [],
                "expected_final": stage_order("nested-dependent")
                + [
                    dict(part="main", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                    dict(part="dependent", step=steps.PRIME),
                ],
                "expected_dirty": [dict(part="nested-dependent", step=steps.PRIME)],
            }
        ),

        self.run_test()


class NestedDependentTest(ReStepOrderTestBase):
    def test_pull_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.PULL,
                "initial_parts": ["nested-dependent"],
                "test_parts": ["nested-dependent"],
                "test_step": steps.PULL,
                "expected_initial": pull_order("nested-dependent"),
                "expected_test": [dict(part="nested-dependent", step=steps.PULL)],
                "expected_test_cleaned": [],
                "expected_final": pull_order("nested-dependent"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_build_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.BUILD,
                "initial_parts": ["nested-dependent"],
                "test_parts": ["nested-dependent"],
                "test_step": steps.PULL,
                "expected_initial": build_order("nested-dependent"),
                "expected_test": [dict(part="nested-dependent", step=steps.PULL)],
                "expected_test_cleaned": [
                    dict(part="nested-dependent", step=steps.BUILD)
                ],
                "expected_final": pull_order("nested-dependent"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_stage_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.STAGE,
                "initial_parts": ["nested-dependent"],
                "test_parts": ["nested-dependent"],
                "test_step": steps.PULL,
                "expected_initial": stage_order("nested-dependent"),
                "expected_test": [dict(part="nested-dependent", step=steps.PULL)],
                "expected_test_cleaned": [
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                ],
                "expected_final": pull_order("nested-dependent"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_prime_repull(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": ["nested-dependent"],
                "test_parts": ["nested-dependent"],
                "test_step": steps.PULL,
                "expected_initial": prime_order("nested-dependent"),
                "expected_test": [dict(part="nested-dependent", step=steps.PULL)],
                "expected_test_cleaned": [
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": prime_order("dependent")
                + [dict(part="nested-dependent", step=steps.PULL)],
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_build_rebuild(self):
        self.set_attributes(
            {
                "initial_step": steps.BUILD,
                "initial_parts": ["nested-dependent"],
                "test_parts": ["nested-dependent"],
                "test_step": steps.BUILD,
                "expected_initial": build_order("nested-dependent"),
                "expected_test": [dict(part="nested-dependent", step=steps.BUILD)],
                "expected_test_cleaned": [],
                "expected_final": build_order("nested-dependent"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_stage_rebuild(self):
        self.set_attributes(
            {
                "initial_step": steps.STAGE,
                "initial_parts": ["nested-dependent"],
                "test_parts": ["nested-dependent"],
                "test_step": steps.BUILD,
                "expected_initial": stage_order("nested-dependent"),
                "expected_test": [dict(part="nested-dependent", step=steps.BUILD)],
                "expected_test_cleaned": [
                    dict(part="nested-dependent", step=steps.STAGE)
                ],
                "expected_final": build_order("nested-dependent"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_prime_rebuild(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": ["nested-dependent"],
                "test_parts": ["nested-dependent"],
                "test_step": steps.BUILD,
                "expected_initial": prime_order("nested-dependent"),
                "expected_test": [dict(part="nested-dependent", step=steps.BUILD)],
                "expected_test_cleaned": [
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": pull_order("nested-dependent")
                + [
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.BUILD),
                ],
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_stage_restage(self):
        self.set_attributes(
            {
                "initial_step": steps.STAGE,
                "initial_parts": ["nested-dependent"],
                "test_parts": ["nested-dependent"],
                "test_step": steps.STAGE,
                "expected_initial": stage_order("nested-dependent"),
                "expected_test": [dict(part="nested-dependent", step=steps.STAGE)],
                "expected_test_cleaned": [],
                "expected_final": stage_order("nested-dependent"),
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_prime_restage(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": ["nested-dependent"],
                "test_parts": ["nested-dependent"],
                "test_step": steps.STAGE,
                "expected_initial": prime_order("nested-dependent"),
                "expected_test": [dict(part="nested-dependent", step=steps.STAGE)],
                "expected_test_cleaned": [
                    dict(part="nested-dependent", step=steps.PRIME)
                ],
                "expected_final": build_order("nested-dependent")
                + [
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.STAGE),
                ],
                "expected_dirty": [],
            }
        )

        self.run_test()

    def test_prime_reprime(self):
        self.set_attributes(
            {
                "initial_step": steps.PRIME,
                "initial_parts": ["nested-dependent"],
                "test_parts": ["nested-dependent"],
                "test_step": steps.PRIME,
                "expected_initial": prime_order("nested-dependent"),
                "expected_test": [dict(part="nested-dependent", step=steps.PRIME)],
                "expected_test_cleaned": [],
                "expected_final": prime_order("nested-dependent"),
                "expected_dirty": [],
            }
        )

        self.run_test()


class DirtyOutdatedTestBase(OrderTestBase):
    def setUp(self):
        super().setUp()

        for directory in ("main", "dependent", "nested-dependent"):
            os.mkdir(directory)
        self.project_config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                parts:
                  main:
                    plugin: dump
                    source: main/
                  dependent:
                    plugin: dump
                    source: dependent/
                    after: [main]
                  nested-dependent:
                    plugin: dump
                    source: nested-dependent/
                    after: [dependent]
                """
            )
        )

        self.dirty_parts = []
        dirty_parts = self.dirty_parts

        original_dirty_report = pluginhandler.PluginHandler.get_dirty_report

        def _fake_dirty_report(self, step):
            nonlocal dirty_parts
            if dict(part=self.name, step=step) in dirty_parts:
                return pluginhandler.DirtyReport(dirty_properties={"forced"})
            return original_dirty_report(self, step)

        patcher = mock.patch.object(
            pluginhandler.PluginHandler, "get_dirty_report", _fake_dirty_report
        )
        patcher.start()
        self.addCleanup(patcher.stop)

        original_clean = pluginhandler.PluginHandler.clean

        def _fake_clean(self, staged_state, primed_state, step):
            nonlocal dirty_parts
            original_clean(self, staged_state, primed_state, step)
            with contextlib.suppress(ValueError):
                dirty_parts.remove(dict(part=self.name, step=step))

        patcher = mock.patch.object(pluginhandler.PluginHandler, "clean", _fake_clean)
        patcher.start()
        self.addCleanup(patcher.stop)

    def make_outdated(self, part_name, step):
        if step == steps.PULL:
            open(os.path.join(part_name, "file"), "w").close()
        else:
            part = self.project_config.parts.get_part(part_name)
            state_file_path = states.get_step_state_file(part.plugin.statedir, step)
            open(state_file_path, "w").close()

    def run_test(self):
        # Prime all parts
        lifecycle.execute(steps.PRIME, self.project_config)

        initial_order = self.get_run_order()
        self.assert_run_order_equal(initial_order, self.expected_initial, "(initial)")

        if self.test_type == "dirty":
            # Make some parts dirty
            self.dirty_parts.extend(self.test_parts)
        else:
            # Make some parts outdated
            for part in self.test_parts:
                self.make_outdated(part["part"], part["step"])

        # Prime again
        lifecycle.execute(steps.PRIME, self.project_config)
        current_order = self.get_run_order()
        test_order = [o for o in current_order if o not in initial_order]
        self.assert_run_order_equal(test_order, self.expected_test, "(test)")

        self.assert_parts_cleaned(initial_order, current_order, [], "(cleaned)")

        self.assert_run_order_equal(
            self.get_run_order(), self.expected_final, "(final)"
        )

        self.assert_parts_dirty([], "(dirty)")


class DirtyTest(DirtyOutdatedTestBase):
    test_type = "dirty"

    def test_main_pull_dirty(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="main", step=steps.PULL)],
                "expected_initial": prime_order(),
                "expected_test": prime_order(),
                "expected_final": prime_order(),
            }
        )

        self.run_test()

    def test_main_build_dirty(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="main", step=steps.BUILD)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="main", step=steps.BUILD),
                    dict(part="main", step=steps.STAGE),
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": prime_order(),
            }
        )

        self.run_test()

    def test_main_stage_dirty(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="main", step=steps.STAGE)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="main", step=steps.STAGE),
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_test_cleaned": [],
                "expected_final": prime_order(),
            }
        )

        self.run_test()

    def test_main_prime_dirty(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="main", step=steps.PRIME)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_test_cleaned": [],
                "expected_final": prime_order(),
            }
        )

        self.run_test()

    def dependent_pull_dirty(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="dependent", step=steps.PULL)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": prime_order("main")
                + [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )

        self.run_test()

    def test_dependent_build_dirty(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="dependent", step=steps.BUILD)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": stage_order("main")
                + [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )

        self.run_test()

    def test_dependent_stage_dirty(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="dependent", step=steps.STAGE)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": stage_order("main")
                + [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )
        self.run_test()

    def test_dependent_prime_dirty(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="dependent", step=steps.PRIME)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": prime_order(),
            }
        )

        self.run_test()

    def test_nested_dependent_pull_dirty(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="nested-dependent", step=steps.PULL)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": prime_order("dependent")
                + [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )

        self.run_test()

    def test_nested_dependendent_build_dirty(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="nested-dependent", step=steps.BUILD)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": pull_order("nested-dependent")
                + [
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )

        self.run_test()

    def test_nested_dependent_stage_dirty(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="nested-dependent", step=steps.STAGE)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": build_order("nested-dependent")
                + [
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )

        self.run_test()

    def test_nested_dependent_prime_dirty(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="nested-dependent", step=steps.PRIME)],
                "expected_initial": prime_order(),
                "expected_test": [dict(part="nested-dependent", step=steps.PRIME)],
                "expected_final": prime_order("nested-dependent"),
            }
        )

        self.run_test()


class OutdatedTest(DirtyOutdatedTestBase):
    test_type = "outdated"

    def test_main_pull_outdated(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="main", step=steps.PULL)],
                "expected_initial": prime_order(),
                "expected_test": prime_order(),
                "expected_final": prime_order(),
            }
        )

        self.run_test()

    def test_main_build_outdated(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="main", step=steps.BUILD)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="main", step=steps.BUILD),
                    dict(part="main", step=steps.STAGE),
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": prime_order(),
            }
        )

        self.run_test()

    def test_main_stage_outdated(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="main", step=steps.STAGE)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="main", step=steps.STAGE),
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_test_cleaned": [],
                "expected_final": prime_order(),
            }
        )

        self.run_test()

    def test_main_prime_outdated(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="main", step=steps.PRIME)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_test_cleaned": [],
                "expected_final": prime_order(),
            }
        )

        self.run_test()

    def dependent_pull_outdated(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="dependent", step=steps.PULL)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": prime_order("main")
                + [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )

        self.run_test()

    def test_dependent_build_outdated(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="dependent", step=steps.BUILD)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": stage_order("main")
                + [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )

        self.run_test()

    def test_dependent_stage_outdated(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="dependent", step=steps.STAGE)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": stage_order("main")
                + [
                    dict(part="dependent", step=steps.PULL),
                    dict(part="dependent", step=steps.BUILD),
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )
        self.run_test()

    def test_dependent_prime_outdated(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="dependent", step=steps.PRIME)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": prime_order(),
            }
        )

        self.run_test()

    def test_nested_dependent_pull_outdated(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="nested-dependent", step=steps.PULL)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": prime_order("dependent")
                + [
                    dict(part="nested-dependent", step=steps.PULL),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )

        self.run_test()

    def test_nested_dependendent_build_outdated(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="nested-dependent", step=steps.BUILD)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": pull_order("nested-dependent")
                + [
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.BUILD),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )

        self.run_test()

    def test_nested_dependent_stage_outdated(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="nested-dependent", step=steps.STAGE)],
                "expected_initial": prime_order(),
                "expected_test": [
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
                "expected_final": build_order("nested-dependent")
                + [
                    dict(part="main", step=steps.PRIME),
                    dict(part="dependent", step=steps.PRIME),
                    dict(part="nested-dependent", step=steps.STAGE),
                    dict(part="nested-dependent", step=steps.PRIME),
                ],
            }
        )

        self.run_test()

    def test_nested_dependent_prime_outdated(self):
        self.set_attributes(
            {
                "test_parts": [dict(part="nested-dependent", step=steps.PRIME)],
                "expected_initial": prime_order(),
                "expected_test": [dict(part="nested-dependent", step=steps.PRIME)],
                "expected_final": prime_order("nested-dependent"),
            }
        )

        self.run_test()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING, Dict, Optional

from snapcraft import formatting_utils
from snapcraft.internal import common, steps

if TYPE_CHECKING:
    from snapcraft.project import Project

    from . import PluginHandler


def get_snapcraft_global_environment(
    project: "Project", include_prime: bool = True
) -> Dict[str, str]:
    """
    Return a global environment for a part.

    include_prime defaults to True for behaviour not to change when using
    core and core18 bases.

    :param Project project: the Snapcraft project instance.
    :param bool include_prime: whether to include the prime directory in
                               the results.
    :returns: a dictionary with environment variables to apply.
    :rtype: dict
    """
    # The triple check is mostly for unit tests support until we completely
    # get rid of project.info.
    if project._snap_meta.name:
        name = project._snap_meta.name
    elif project.info is not None and project.info.name:
        name = project.info.name
    else:
        name = ""

    if project._snap_meta.version:
        version = project._snap_meta.version
    elif project.info is not None and project.info.version:
        version = project.info.version
    else:
        version = ""

    if project._snap_meta.grade:
        grade = project._snap_meta.grade
    elif project.info is not None and project.info.grade:
        grade = project.info.grade
    else:
        grade = ""

    return {
        "SNAPCRAFT_ARCH_TRIPLET": project.arch_triplet,
        "SNAPCRAFT_EXTENSIONS_DIR": common.get_extensionsdir(),
        "SNAPCRAFT_PARALLEL_BUILD_COUNT": str(project.parallel_build_count),
        "SNAPCRAFT_PRIME": project.prime_dir,
        "SNAPCRAFT_PROJECT_NAME": name,
        "SNAPCRAFT_PROJECT_VERSION": version,
        "SNAPCRAFT_PROJECT_DIR": project._project_dir,
        "SNAPCRAFT_PROJECT_GRADE": grade,
        "SNAPCRAFT_STAGE": project.stage_dir,
        "SNAPCRAFT_TARGET_ARCH": project.target_arch,
    }


def get_snapcraft_part_directory_environment(
    part: "PluginHandler", *, step: Optional[steps.Step] = None
) -> Dict[str, str]:
    env = {
        "SNAPCRAFT_PART_SRC": part.part_source_dir,
        "SNAPCRAFT_PART_SRC_WORK": part.part_source_work_dir,
    }

    if step is None or step == steps.BUILD:
        env.update(
            {
                "SNAPCRAFT_PART_BUILD": part.part_build_dir,
                "SNAPCRAFT_PART_BUILD_WORK": part.part_build_work_dir,
                "SNAPCRAFT_PART_INSTALL": part.part_install_dir,
            }
        )

    return env


def get_snapcraft_part_environment(
    part: "PluginHandler", *, step: steps.Step
) -> Dict[str, str]:
    """Return Snapcraft provided part environment."""
    part_environment = get_snapcraft_global_environment(
        part._project, include_prime=step == steps.PRIME
    )
    part_environment.update(get_snapcraft_part_directory_environment(part, step=step))

    paths = [part.part_install_dir, part._project.stage_dir]

    bin_paths = list()
    for path in paths:
        bin_paths.extend(common.get_bin_paths(root=path, existing_only=True))

    if bin_paths:
        bin_paths.append("$PATH")
        part_environment["PATH"] = formatting_utils.combine_paths(
            paths=bin_paths, prepend="", separator=":"
        )

    include_paths = list()
    for path in paths:
        include_paths.extend(common.get_include_paths(path, part._project.arch_triplet))

    if include_paths:
        for envvar in ["CPPFLAGS", "CFLAGS", "CXXFLAGS"]:
            part_environment[envvar] = formatting_utils.combine_paths(
                paths=include_paths, prepend="-isystem", separator=" "
            )

    library_paths = list()
    for path in paths:
        library_paths.extend(common.get_library_paths(path, part._project.arch_triplet))

    if library_paths:
        part_environment["LDFLAGS"] = formatting_utils.combine_paths(
            paths=library_paths, prepend="-L", separator=" "
        )

    pkg_config_paths = common.get_pkg_config_paths(path, part._project.arch_triplet)
    if pkg_config_paths:
        part_environment["PKG_CONFIG_PATH"] = formatting_utils.combine_paths(
            pkg_config_paths, prepend="", separator=":"
        )

    return part_environment



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import pathlib
import shutil
import subprocess
from collections import OrderedDict
from copy import deepcopy

import snapcraft
import snapcraft.internal.sources
from snapcraft import yaml_utils
from snapcraft.file_utils import rmtree
from snapcraft.internal.meta import _version
from snapcraft.internal.remote_build import errors
from snapcraft.project import Project

logger = logging.getLogger(__name__)


class WorkTree:
    """Manages tree for remote-build project."""

    def __init__(
        self, worktree_dir: str, project: Project, package_all_sources=False
    ) -> None:
        """Create remote-build WorkTree.

        :param str worktree_dir: Directory to use for working tree.
        :param Project project: Snapcraft project.
        :param bool package_all_sources: Package all sources instead of
               the default of local-only sources.
        """
        self._project = project
        self._package_all_sources = package_all_sources

        # Get snapcraft yaml (as OrderedDict).
        self._snapcraft_config = (
            self._project.info.get_raw_snapcraft()
        )  # type: OrderedDict

        # Working copy.
        self._prepared_snapcraft_config = (
            self._project.info.get_raw_snapcraft()
        )  # type: OrderedDict

        # Working tree base directory.
        self._base_dir = worktree_dir

        # Initialize cache for sources.
        self._cache_dir = os.path.join(self._base_dir, "cache")
        os.makedirs(self._cache_dir, exist_ok=True)

        # Initialize clean repo to ship to remote builder.
        self._repo_dir = os.path.join(self._base_dir, "repo")
        if os.path.exists(self._repo_dir):
            rmtree(self._repo_dir)

        self._repo_sources_dir = os.path.join(self._repo_dir, "sources")
        self._repo_snap_dir = os.path.join(self._repo_dir, "snap")

    def _get_part_source_handler(self, part_name: str, source: str, source_dir: str):
        """Get an instance for a part's source handler.

        :param str part_name: Name of part.
        :param str source: Source URL.
        :param str source_dir: Directory to pull sources to.
        :return: Source handler instanced configured with source and source_dir.
        """

        # The actual source may be a selector, override it.
        part_config = self._snapcraft_config["parts"][part_name]
        part_config["source"] = source
        source_type = part_config.get("source-type")

        handler_class = snapcraft.internal.sources.get_source_handler(
            source, source_type=source_type
        )
        return handler_class(
            source,
            source_dir,
            source_checksum=part_config.get("source-checksum"),
            source_branch=part_config.get("source-branch"),
            source_tag=part_config.get("source-tag"),
            source_depth=part_config.get("source-depth"),
            source_commit=part_config.get("source-commit"),
        )

    def _get_part_cache_dir(self, part_name: str, selector=None) -> str:
        """Get source cache directory for part_name.

        Takes into account source selector, if provided, as parts using source
        selectors will require multiple source directories.

        :param str part_name: Name of part.
        :param str selector: Source selector, if any (e.g. "on amd64").
        :return: Directory path to store source.
        """
        if selector:
            return os.path.join(self._cache_dir, part_name, selector)

        return os.path.join(self._cache_dir, part_name)

    def _get_part_tarball_path(self, part_name: str, selector=None) -> str:
        """Get path to archive the specified part's source tarball in the repo.

        :param str part_name: Name of part.
        :param str selector: Source selector, if any (e.g. "on amd64").
        :return: File path for tarball.
        """
        tarball_filename = part_name + ".tar.gz"
        if selector:
            return os.path.join(
                self._repo_sources_dir, part_name, selector, tarball_filename
            )
        return os.path.join(self._repo_sources_dir, part_name, tarball_filename)

    def _archive_part_sources(self, part_name: str, selector=None):
        """Archive sources at source_dir into archive_path.

        :param str part_name: Name of part.
        :param str selector: Source selector, if any (e.g. "on amd64").
        :return: Relative path to archive within repository.
        """
        source_path = self._get_part_cache_dir(part_name, selector)
        archive_path = self._get_part_tarball_path(part_name, selector)

        logger.debug("creating source archive: {}".format(archive_path))

        os.makedirs(os.path.split(archive_path)[0], exist_ok=True)
        command = ["tar", "czf", archive_path, "-C", source_path, "."]
        subprocess.check_call(command)
        relpath = os.path.relpath(archive_path, self._repo_dir)

        # Ensure Linux path formatting is used.
        return pathlib.Path(relpath).as_posix()

    def _pull_source(self, part_name: str, source: str, selector=None) -> str:
        """Pull source_url for part to source_dir. Returns source.

        :param str part_name: Name of part.
        :param str source: Source URL to pull.
        :param str selector: Source selector, if any (e.g. "on amd64").
        :return: Relative path to archive within repository.
        """
        download_dir = self._get_part_cache_dir(part_name, selector)
        source_handler = self._get_part_source_handler(part_name, source, download_dir)

        if selector:
            print_name = "{} [selector='{}']".format(part_name, selector)
        else:
            print_name = part_name

        # Skip non-local sources (the remote builder can fetch those directly),
        # unless configured to package all sources.
        is_local_source = isinstance(
            source_handler, snapcraft.internal.sources.Local
        ) or (
            isinstance(source_handler, snapcraft.internal.sources.Git)
            and source_handler.is_local()
        )
        if not self._package_all_sources and not is_local_source:
            logger.debug("passing through source for {}: {}".format(print_name, source))
            return source

        logger.info("Packaging sources for {}...".format(print_name))

        # Remove existing cache directory (if exists)
        if os.path.exists(download_dir):
            rmtree(download_dir)

        # Pull sources, but create directory first if part
        # is configured to use the `dump` plugin.
        if self._snapcraft_config["parts"][part_name].get("plugin") == "dump":
            os.makedirs(download_dir, exist_ok=True)

        source_handler.pull()

        # Create source archive.
        return self._archive_part_sources(part_name, selector)

    def _process_part_sources(
        self, part_name: str, part_config: OrderedDict
    ) -> OrderedDict:
        """Add part sources to tree.

        :param str part_name: Name of part.
        :param OrderedDict part_config: Part configuration for specified part.
        :return: Modified part configuration for prepared snapcraft.yaml.
        """
        logger.debug("processing {}: {}".format(part_name, part_config))

        # We will be modifying part_config, copy it.
        part_config = deepcopy(part_config)

        if "source" not in part_config:
            logger.debug("no source specified, skipping...")
            return part_config

        source = part_config["source"]
        source_modified = False
        if isinstance(source, list):
            # Handle advanced grammar case for `source` by parsing an
            # array of {<selector>: <source>} OrderedDicts. Example:
            # source:
            # - on amd64: http://archive.ubuntu.com/foo_amd64.deb
            # - on arm64: http://archive.ubuntu.com/foo_arm64.deb
            # - on i386: http://archive.ubuntu.com/foo_i386.deb
            # Results in:
            # [
            #   {'on amd64': 'http://archive.ubuntu.com/foo_amd64.deb'},
            #   {'on arm64': 'http://archive.ubuntu.com/foo_arm64.deb'},
            #   {'on i386': 'http://archive.ubuntu.com/foo_i386.deb'},
            # ]
            part_config["source"] = []
            for s in source:
                sinfo = OrderedDict()  # type: OrderedDict
                # Generally speaking, there will only be one key in each
                # dict (e.g. "on amd64"), but we iterate regardless.
                for selector, selector_source in s.items():
                    sinfo[selector] = self._pull_source(
                        part_name, selector_source, selector=selector
                    )
                    if sinfo[selector] != selector_source:
                        source_modified = True

                part_config["source"].append(sinfo)
        else:
            part_config["source"] = self._pull_source(part_name, source)
            if part_config["source"] != source:
                source_modified = True

        if source_modified:
            # Strip source attributes no longer relevant.
            for key in list(part_config.keys()):
                if key.startswith("source-") and key != "source-subdir":
                    part_config.pop(key)

            # It's now an archive, set it explicitly.
            part_config["source-type"] = "tar"

        return part_config

    def _copy_assets(self):
        """Copy assets directly to repository, if exists."""
        assets_dir = self._project._get_snapcraft_assets_dir()
        logger.debug("assets expected at: {}".format(assets_dir))

        # Assets directory exists, but may not actually contain assets.
        # The snapd project is an example of this.  The sanity checker
        # will warn to the user when the project is built.
        if os.path.exists(assets_dir):
            logger.debug(
                "copying assets: {} -> {}".format(assets_dir, self._repo_snap_dir)
            )

            # Remove existing assets, and update copy.
            if os.path.exists(self._repo_snap_dir):
                rmtree(self._repo_snap_dir)
            shutil.copytree(assets_dir, self._repo_snap_dir)
        else:
            os.makedirs(self._repo_snap_dir, exist_ok=True)

        # Copy plugins if not already inside assets directory (and exists).
        plugins_dir = self._project._get_local_plugins_dir()
        if plugins_dir.startswith(assets_dir) or not os.path.exists(plugins_dir):
            return

        self._repo_plugins_dir = os.path.join(self._repo_snap_dir, "plugins")
        logger.debug(
            "copying local plugins: {} -> {}".format(
                plugins_dir, self._repo_plugins_dir
            )
        )
        shutil.copytree(plugins_dir, self._repo_plugins_dir)

    def _set_prepared_project_version(self):
        """Set project version to handle `version: git`
        This and version-script expect to run in project root (".").
        """
        # version-script is deprecated, remote-build won't support it.
        if "version-script" in self._snapcraft_config:
            raise errors.UnsupportedVersionScriptError()

        version = self._snapcraft_config.get("version")

        # Nothing to do here if no version is set.
        if version is None:
            return

        # Determine git version, if `version: git`.
        if version == "git":
            try:
                version = _version.get_version(version, version_script=None)
            except FileNotFoundError:
                raise errors.GitNotFoundVersionError()

        self._prepared_snapcraft_config["version"] = version

    def prepare_repository(self) -> str:
        """Prepare source tree for launchpad build. Returns repo directory."""
        # Copy project assets.
        self._copy_assets()

        # Create sources directory for source archives.
        os.makedirs(self._repo_sources_dir, exist_ok=True)

        # Process each part with sources.
        for part_name, part_config in self._snapcraft_config["parts"].items():
            part_config = self._process_part_sources(part_name, part_config)
            self._prepared_snapcraft_config["parts"][part_name] = part_config

        # Set version.
        self._set_prepared_project_version()

        # Write updated snapcraft yaml config.
        snapcraft_yaml_path = os.path.join(self._repo_snap_dir, "snapcraft.yaml")
        with open(snapcraft_yaml_path, "w") as f:
            yaml_utils.dump(self._prepared_snapcraft_config, stream=f)

        return self._repo_dir



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

from testtools.matchers import Contains, Equals

from snapcraft import storeapi

from . import FakeStoreCommandsBaseTestCase


class ListTracksCommandTestCase(FakeStoreCommandsBaseTestCase):
    def test_list_tracks_without_snap_raises_exception(self):
        result = self.run_command(["list-tracks"])

        self.assertThat(result.exit_code, Equals(2))
        self.assertThat(result.output, Contains("Usage:"))

    def test_list_tracks_without_login_must_ask(self):
        self.fake_store_get_snap_channel_map.mock.side_effect = [
            storeapi.http_clients.errors.InvalidCredentialsError("error"),
            self.channel_map,
        ]

        result = self.run_command(
            ["list-tracks", "snap-test"], input="user@example.com\nsecret\n"
        )
        self.assertThat(
            result.output, Contains("You are required to login before continuing.")
        )

    def test_list_tracks(self):
        result = self.run_command(["list-tracks", "snap-test"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                dedent(
                    """\
            Name    Status    Creation-Date         Version-Pattern
            latest  active    -                     -
            2.0     default   2019-10-17T14:11:59Z  2\\.*
            """
                )
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2017 Marius Gripsgard (mariogrip@ubuntu.com)
# Copyright (C) 2016-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This rust plugin is useful for building rust based parts.

Rust uses cargo to drive the build.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - rust-channel
      (string)
      select rust channel (stable, beta, nightly)
    - rust-revision
      (string)
      select rust version, this is preferred over rust-channel
    - rust-features
      (list of strings)
      Features used to build optional dependencies

If a rust-toolchain file is found, the toolchain specified there will
be used unless rust-channel or rust-revision are set which will
override the rust-toolchain file. If neither rust-toolchain exists nor
rust-channel or rust-revision are set, the latest stable toolchain
will be used.
"""

import collections
import logging
import os
from contextlib import suppress
from pathlib import Path
from typing import List, Optional

import toml

from snapcraft import file_utils, shell_utils, sources
from snapcraft.internal import errors
from snapcraft.plugins.v1 import PluginV1

_RUSTUP = "https://sh.rustup.rs/"
logger = logging.getLogger(__name__)


class RustPlugin(PluginV1):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["rust-channel"] = {
            "type": "string",
            "enum": ["stable", "beta", "nightly"],
        }
        schema["properties"]["rust-revision"] = {"type": "string"}
        schema["properties"]["rust-features"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_pull_properties(cls):
        return ["rust-revision", "rust-channel"]

    @classmethod
    def get_build_properties(cls):
        return ["rust-features"]

    def __init__(self, name, options, project):
        super().__init__(name, options, project)
        self.build_packages.extend(["gcc", "git", "curl", "file"])
        self._rustup_dir = os.path.expanduser(os.path.join("~", ".rustup"))
        self._cargo_dir = os.path.expanduser(os.path.join("~", ".cargo"))
        self._rustup_cmd = os.path.join(self._cargo_dir, "bin", "rustup")
        self._cargo_cmd = os.path.join(self._cargo_dir, "bin", "cargo")
        self._rustc_cmd = os.path.join(self._cargo_dir, "bin", "rustc")
        self._rustdoc_cmd = os.path.join(self._cargo_dir, "bin", "rustdoc")

        self._manifest = collections.OrderedDict()

        if self.options.source_subdir:
            self.source_path = Path(self.sourcedir, self.options.source_subdir)
        else:
            self.source_path = Path(self.sourcedir)

    def enable_cross_compilation(self):
        # The logic is applied transparently through internal
        # rust tooling.
        pass

    def pull(self):
        super().pull()
        self._fetch_rustup()
        self._fetch_rust()
        self._fetch_cargo_deps()

    def _fetch_rustup(self):
        # if rustup-init has already been done, we can skip this.
        if os.path.exists(self._rustup_cmd):
            return

        # Download rustup-init.
        os.makedirs(self._rustup_dir, exist_ok=True)
        rustup_init_cmd = os.path.join(self._rustup_dir, "rustup.sh")
        sources.Script(_RUSTUP, self._rustup_dir).download(filepath=rustup_init_cmd)

        # Basic options:
        # -y: assume yes
        # --no-modify-path: do not modify bashrc
        options = ["-y", "--no-modify-path", "--profile=minimal"]
        if self._get_toolchain() is not None:
            options.extend(["--default-toolchain", "none"])

        # Fetch rust
        self.run([rustup_init_cmd] + options, env=self._build_env())

    def _fetch_rust(self) -> None:
        toolchain = self._get_toolchain()

        # We only do this when one of the snapcraft properties are set,
        # the reason for this is faster iterations when changing any
        # of the toolchain related properties.
        # When changing the rust-toolchain file a re-pull will be triggered
        # due to source changes and the right thing will happen.
        if toolchain is not None:
            # https://rust-lang-nursery.github.io/edition-guide/rust-2018/rustup-for-managing-rust-versions.html
            self.run([self._rustup_cmd, "install", toolchain], env=self._build_env())

        # Add the appropriate target cross compilation target if necessary.
        # https://github.com/rust-lang/rustup.rs/blob/master/README.md#cross-compilation
        if self.project.is_cross_compiling:
            add_target_cmd = [self._rustup_cmd, "target", "add"]
            if toolchain is not None:
                add_target_cmd.extend(["--toolchain", toolchain])
            add_target_cmd.append(self._get_target())

            self.run(add_target_cmd, env=self._build_env())

    def _project_uses_cargo_lock(self) -> bool:
        return Path(self.source_path, "Cargo.lock").exists()

    def _fetch_cargo_deps(self):
        if self.options.source_subdir:
            sourcedir = os.path.join(self.sourcedir, self.options.source_subdir)
        else:
            sourcedir = self.sourcedir

        fetch_cmd = [
            self._cargo_cmd,
            "fetch",
            "--manifest-path",
            os.path.join(sourcedir, "Cargo.toml"),
        ]

        # Use Cargo.lock, if available.
        if self._project_uses_cargo_lock():
            fetch_cmd.append("--locked")

        toolchain = self._get_toolchain()
        if toolchain is not None:
            fetch_cmd.insert(1, "+{}".format(toolchain))
        self.run(fetch_cmd, env=self._build_env())

    def _get_target(self) -> str:
        # Cf. rustc --print target-list
        targets = {
            "armhf": "armv7-{}-{}eabihf",
            "arm64": "aarch64-{}-{}",
            "i386": "i686-{}-{}",
            "amd64": "x86_64-{}-{}",
            "ppc64el": "powerpc64le-{}-{}",
            "s390x": "s390x-{}-{}",
        }
        rust_target = targets.get(self.project.deb_arch)
        if not rust_target:
            raise errors.SnapcraftEnvironmentError(
                "{!r} is not supported as a target architecture ".format(
                    self.project.deb_arch
                )
            )
        return rust_target.format("unknown-linux", "gnu")

    def _project_uses_workspace(self) -> bool:
        cargo_toml_path = Path(self.builddir, "Cargo.toml")

        try:
            config = open(cargo_toml_path).read()
        except FileNotFoundError:
            raise errors.SnapcraftPluginAssertionError(
                name=self.name, reason="missing required Cargo.toml in source directory"
            )

        return "workspace" in toml.loads(config)

    def _install_workspace_artifacts(self) -> None:
        """Install workspace artifacts."""
        # Find artifacts in release directory.
        release_dir = Path(self.builddir, "target", "release")

        # Install binaries to bin/.
        bins_dir = Path(self.installdir, "bin")
        bins_dir.mkdir(parents=True, exist_ok=True)

        # Install shared objects to usr/lib/<arch-triplet>.
        # TODO: Dynamic library support needs to be properly added.
        # Although we install libraries if we find them, they are most
        # likely going to be missing dependencies, e.g.:
        # /home/ubuntu/.cargo/toolchains/stable-x86_64-unknown-linux-gnu/lib/libstd-fae576517123aa4e.so
        libs_dir = Path(self.installdir, "usr", "lib", self.project.arch_triplet)
        libs_dir.mkdir(parents=True, exist_ok=True)

        # Cargo build marks binaries and shared objects executable...
        # Search target directory to get these and install them to the
        # correct location (*.so to lib directory, otherwise bin directory).
        executables = [
            i
            for i in release_dir.iterdir()
            if os.path.isfile(i) and os.access(i, os.X_OK)
        ]
        for exe_path in executables:
            if exe_path.name.endswith(".so"):
                file_utils.link_or_copy(exe_path.as_posix(), libs_dir.as_posix())
            else:
                file_utils.link_or_copy(exe_path.as_posix(), bins_dir.as_posix())

    def build(self):
        super().build()

        # Write a minimal config.
        self._write_cargo_config()

        uses_workspaces = self._project_uses_workspace()
        if uses_workspaces:
            # This is a bit ugly because `cargo install` does not yet support
            # workspaces.  Alternatively, there is a perhaps better option
            # to use `cargo-build --out-dir`, but `--out-dir` is considered
            # unstable and unavailable for use yet on the stable channel.
            install_cmd = [self._cargo_cmd, "build", "--release"]
        else:
            install_cmd = [
                self._cargo_cmd,
                "install",
                "--path",
                self.builddir,
                "--root",
                self.installdir,
                "--force",
            ]

            if self._project_uses_cargo_lock():
                install_cmd.append("--locked")

        toolchain = self._get_toolchain()
        if toolchain is not None:
            install_cmd.insert(1, "+{}".format(toolchain))

        # Even though this is mostly harmless when not cross compiling
        # the flag is in place to avoid a situation where an earlier
        # version of the toolchain is used.
        if self.project.is_cross_compiling:
            install_cmd.extend(["--target", self._get_target()])

        if self.options.rust_features:
            install_cmd.append("--features")
            install_cmd.append(" ".join(self.options.rust_features))

        # build and install.
        self.run(install_cmd, env=self._build_env(), cwd=self.builddir)

        # Finally, record.
        self._record_manifest()

        if uses_workspaces:
            # We need to install the workspace artifacts as a workaround until
            # `cargo build` supports `out-dir` in "stable".
            self._install_workspace_artifacts()

    def _build_env(self):
        env = os.environ.copy()

        env.update(dict(RUSTUP_HOME=self._rustup_dir, CARGO_HOME=self._cargo_dir))

        rustflags = self._get_rustflags()
        if rustflags:
            string_fmt = " ".join(["{}".format(i) for i in rustflags]).strip()
            env.update(RUSTFLAGS=string_fmt)

        return env

    def _get_toolchain(self) -> Optional[str]:
        """Return the toolchain to use.

        If a rust-toolchain file is present, None will be returned.
        If a rust-toolchain file is not present and neither rust-version
        nor rust-channel are set, then stable will be returned.
        """
        if self.options.source_subdir:
            sourcedir = os.path.join(self.sourcedir, self.options.source_subdir)
        else:
            sourcedir = self.sourcedir
        rust_toolchain_path = os.path.join(sourcedir, "rust-toolchain")

        if self.options.rust_revision:
            toolchain = self.options.rust_revision
        elif self.options.rust_channel:
            toolchain = self.options.rust_channel
        elif not os.path.exists(rust_toolchain_path):
            toolchain = "stable"
        else:
            toolchain = None

        return toolchain

    def _get_linker(self) -> str:
        arch_triplet = self.project.arch_triplet
        if arch_triplet == "i386-linux-gnu":
            return "i686-linux-gnu-gcc"
        else:
            return "{}-gcc".format(arch_triplet)

    def _write_cargo_config(self, cargo_config_path: Optional[str] = None) -> None:
        if cargo_config_path is None:
            cargo_config_path = os.path.join(self.builddir, ".cargo", "config")
        if os.path.dirname(cargo_config_path):
            os.makedirs(os.path.dirname(cargo_config_path), exist_ok=True)

        target = {self._get_target(): dict(linker=self._get_linker())}
        config = dict(
            arch_triplet=self.project.arch_triplet,
            jobs=self.parallel_build_count,
            rustc_cmd=self._rustc_cmd,
            rustdoc_cmd=self._rustdoc_cmd,
            target=target,
        )

        # Cf. http://doc.crates.io/config.html
        with open(cargo_config_path, "w") as toml_config_file:
            toml.dump(config, toml_config_file)

    def _get_rustflags(self) -> List[str]:
        ldflags = shell_utils.getenv("LDFLAGS")
        rustldflags = []
        flags = {flag for flag in ldflags.split(" ") if flag}
        for flag in flags:
            rustldflags.extend(["-C", "link-arg={}".format(flag)])

        if self.project.is_cross_compiling:
            rustldflags.extend(["-C", "linker={}".format(self._get_linker())])

        return rustldflags

    def _record_manifest(self):
        self._manifest["rustup-version"] = self.run_output(
            [self._rustup_cmd, "--version"], env=self._build_env()
        )
        toolchain = self._get_toolchain()
        if toolchain is not None:
            toolchain_option = "+{}".format(toolchain)
            self._manifest["rustc-version"] = self.run_output(
                [self._rustc_cmd, toolchain_option, "--version"], env=self._build_env()
            )
            self._manifest["cargo-version"] = self.run_output(
                [self._cargo_cmd, toolchain_option, "--version"], env=self._build_env()
            )
        else:
            self._manifest["rustc-version"] = self.run_output(
                [self._rustc_cmd, "--version"], env=self._build_env()
            )
            self._manifest["cargo-version"] = self.run_output(
                [self._cargo_cmd, "--version"], env=self._build_env()
            )
        with suppress(FileNotFoundError, IsADirectoryError):
            with open(os.path.join(self.builddir, "Cargo.lock")) as lock_file:
                self._manifest["cargo-lock-contents"] = lock_file.read()

    def get_manifest(self):
        return self._manifest



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import textwrap

import testscenarios
from testtools.matchers import Equals

from snapcraft.extractors import ExtractedMetadata, _errors, appstream
from tests import unit


def _create_desktop_file(desktop_file_path, icon: str = None) -> None:
    dir_name = os.path.dirname(desktop_file_path)
    if not os.path.exists(dir_name):
        os.makedirs(dir_name)
    f = open(desktop_file_path, "w")
    print("[Desktop Entry]", file=f)
    if icon:
        print("Icon={}".format(icon), file=f)
    f.close()


class TestAppstream:

    scenarios = testscenarios.multiply_scenarios(
        [
            (
                "summary",
                {
                    "key": "summary",
                    "attributes": {},
                    "param_name": "summary",
                    "value": "test-summary",
                    "expect": "test-summary",
                },
            ),
            (
                "description",
                {
                    "key": "description",
                    "attributes": {},
                    "param_name": "description",
                    "value": "test-description",
                    "expect": "test-description",
                },
            ),
            (
                "local icon with non-existing file",
                {
                    "key": "icon",
                    "attributes": {"type": "local"},
                    "param_name": "icon",
                    "value": "/test/path",
                    "expect": None,
                },
            ),
            (
                "local icon with existing file",
                {
                    "key": "icon",
                    "attributes": {"type": "local"},
                    "param_name": "icon",
                    "value": "/icon.png",
                    "expect": "/icon.png",
                },
            ),
            (
                "common id",
                {
                    "key": "id",
                    "attributes": {},
                    "param_name": "common_id",
                    "value": "test-id",
                    "expect": "test-id",
                },
            ),
            (
                "title",
                {
                    "key": "name",
                    "attributes": {},
                    "param_name": "title",
                    "value": "test-title",
                    "expect": "test-title",
                },
            ),
        ],
        [
            ("metainfo", {"file_extension": "metainfo.xml"}),
            ("appdata", {"file_extension": "appdata.xml"}),
        ],
    )

    def test(
        self, tmp_work_path, file_extension, key, attributes, param_name, value, expect
    ):
        file_name = f"foo.{file_extension}"
        attributes = " ".join(
            '{attribute_name}="{attribute_value}"'.format(
                attribute_name=attribute, attribute_value=attributes[attribute]
            )
            for attribute in attributes
        )
        with open(file_name, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                <?xml version="1.0" encoding="UTF-8"?>
                <component>
                  <{key} {attributes}>{value}</{key}>
                </component>""".format(
                        key=key, value=value, attributes=attributes
                    )
                )
            )

        open("icon.png", "w").close()
        kwargs = {param_name: expect}
        expected = ExtractedMetadata(**kwargs)

        assert appstream.extract(file_name, workdir=".") == expected


# See LP #1814898 for a description of possible fallbacks
class AppstreamIconsTestCase(unit.TestCase):
    def _create_appstream_file(self, icon: str = None, icon_type: str = "local"):
        with open("foo.appdata.xml", "w") as f:
            if icon:
                f.write(
                    textwrap.dedent(
                        """\
                    <?xml version="1.0" encoding="UTF-8"?>
                    <component>
                      <launchable type="desktop-id">my.app.desktop</launchable>
                      <icon type="{}">{}</icon>
                    </component>""".format(
                            icon_type, icon
                        )
                    )
                )
            else:
                f.write(
                    textwrap.dedent(
                        """\
                    <?xml version="1.0" encoding="UTF-8"?>
                    <component>
                      <launchable type="desktop-id">my.app.desktop</launchable>
                    </component>"""
                    )
                )

    def _create_index_theme(self, theme: str):
        # TODO: populate index.theme
        dir_name = os.path.join("usr", "share", "icons", theme)
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
        open(os.path.join(dir_name, "index.theme"), "w").close()

    def _create_icon_file(self, theme: str, size: str, filename: str) -> None:
        dir_name = os.path.join("usr", "share", "icons", theme, size, "apps")
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
        open(os.path.join(dir_name, filename), "w").close()

    def _expect_icon(self, icon):
        expected = ExtractedMetadata(icon=icon)
        actual = appstream.extract("foo.appdata.xml", workdir=".")
        self.assertThat(actual.get_icon(), Equals(expected.get_icon()))

    def test_appstream_NxN_size_not_int_is_skipped(self):
        self._create_appstream_file(icon="icon", icon_type="stock")
        dir_name = os.path.join("usr", "share", "icons", "hicolor", "NxN")
        os.makedirs(dir_name)
        self._expect_icon(None)

    def test_appstream_index_theme_is_not_confused_for_size(self):
        self._create_appstream_file(icon="icon", icon_type="stock")
        self._create_index_theme("hicolor")
        self._expect_icon(None)

    def test_appstream_stock_icon_exists_png(self):
        self._create_appstream_file(icon="icon", icon_type="stock")
        self._create_icon_file("hicolor", "48x48", "icon.png")
        self._create_icon_file("hicolor", "64x64", "icon.png")
        self._expect_icon("usr/share/icons/hicolor/64x64/apps/icon.png")

    def test_appstream_stock_icon_not_exist(self):
        self._create_appstream_file(icon="missing", icon_type="stock")
        self._expect_icon(None)

    def test_appstream_no_icon_no_fallback(self):
        self._create_appstream_file()
        self._expect_icon(None)

    def test_appstream_local_icon_exists(self):
        self._create_appstream_file(icon="/icon.png")
        open("icon.png", "w").close()
        self._expect_icon("/icon.png")

    def test_appstream_local_icon_not_exist_no_fallback(self):
        self._create_appstream_file(icon="/missing.png")
        self._expect_icon(None)

    def test_appstream_local_icon_not_absolute_no_fallback(self):
        self._create_appstream_file(icon="foo")
        self._expect_icon(None)

    def test_appstream_no_icon_desktop_fallback_no_icon(self):
        self._create_appstream_file()
        _create_desktop_file("usr/share/applications/my.app.desktop")
        self._expect_icon(None)

    def test_appstream_no_icon_desktop_fallback_icon_not_exist(self):
        self._create_appstream_file()
        _create_desktop_file(
            "usr/share/applications/my.app.desktop", icon="/missing.png"
        )
        self._expect_icon("/missing.png")

    def test_appstream_no_icon_desktop_fallback_icon_exists(self):
        self._create_appstream_file()
        _create_desktop_file("usr/share/applications/my.app.desktop", icon="/icon.png")
        open("icon.png", "w").close()
        self._expect_icon("/icon.png")

    def test_appstream_no_icon_theme_fallback_png(self):
        self._create_appstream_file()
        _create_desktop_file("usr/share/applications/my.app.desktop", icon="icon")
        self._create_icon_file("hicolor", "scalable", "icon.svg")
        self._create_icon_file("hicolor", "48x48", "icon.png")
        self._create_icon_file("hicolor", "64x64", "icon.png")
        self._expect_icon("usr/share/icons/hicolor/64x64/apps/icon.png")

    def test_appstream_no_icon_theme_fallback_xpm(self):
        self._create_appstream_file()
        _create_desktop_file("usr/share/applications/my.app.desktop", icon="icon")
        self._create_icon_file("hicolor", "scalable", "icon.svg")
        self._create_icon_file("hicolor", "48x48", "icon.png")
        self._create_icon_file("hicolor", "64x64", "icon.xpm")
        self._expect_icon("usr/share/icons/hicolor/64x64/apps/icon.xpm")

    def test_appstream_no_icon_theme_fallback_svg(self):
        self._create_appstream_file()
        _create_desktop_file("usr/share/applications/my.app.desktop", icon="icon")
        self._create_icon_file("hicolor", "scalable", "icon.svg")
        self._expect_icon("usr/share/icons/hicolor/scalable/apps/icon.svg")

    def test_appstream_no_icon_theme_fallback_svgz(self):
        self._create_appstream_file()
        _create_desktop_file("usr/share/applications/my.app.desktop", icon="icon")
        self._create_icon_file("hicolor", "scalable", "icon.svgz")
        self._expect_icon("usr/share/icons/hicolor/scalable/apps/icon.svgz")


class AppstreamTest(unit.TestCase):
    def test_appstream_with_ul(self):
        file_name = "snapcraft.appdata.xml"
        content = textwrap.dedent(
            """\
            <?xml version="1.0" encoding="utf-8"?>
            <component type="desktop">
              <id>io.snapcraft.snapcraft</id>
              <metadata_license>CC0-1.0</metadata_license>
              <project_license>GPL-3.0</project_license>
              <name>snapcraft</name>
              <name xml:lang="es">snapcraft</name>
              <summary>Create snaps</summary>
              <summary xml:lang="es">Crea snaps</summary>
              <description>
                <p>Command Line Utility to create snaps.</p>
                <p xml:lang="es">Aplicativo de línea de comandos para crear snaps.</p>
                <p>Features:</p>
                <p xml:lang="es">Funciones:</p>
                <ul>
                  <li>Build snaps.</li>
                  <li xml:lang="es">Construye snaps.</li>
                  <li>Publish snaps to the store.</li>
                  <li xml:lang="es">Publica snaps en la tienda.</li>
                </ul>
              </description>
              <provides>
                <binary>snapcraft</binary>
              </provides>
            </component>
        """
        )

        with open(file_name, "w") as f:
            print(content, file=f)

        metadata = appstream.extract(file_name, workdir=".")

        self.assertThat(metadata.get_summary(), Equals("Create snaps"))
        self.assertThat(
            metadata.get_description(),
            Equals(
                textwrap.dedent(
                    """\
            Command Line Utility to create snaps.

            Features:

            - Build snaps.
            - Publish snaps to the store."""
                )
            ),
        )

    def test_appstream_with_ol(self):
        file_name = "snapcraft.appdata.xml"
        content = textwrap.dedent(
            """\
            <?xml version="1.0" encoding="utf-8"?>
            <component type="desktop">
              <id>io.snapcraft.snapcraft</id>
              <metadata_license>CC0-1.0</metadata_license>
              <project_license>GPL-3.0</project_license>
              <name>snapcraft</name>
              <name xml:lang="es">snapcraft</name>
              <summary>Create snaps</summary>
              <summary xml:lang="es">Crea snaps</summary>
              <description>
                <p>Command Line Utility to create snaps.</p>
                <p xml:lang="es">Aplicativo de línea de comandos para crear snaps.</p>
                <p>Features:</p>
                <p xml:lang="es">Funciones:</p>
                <ol>
                  <li>Build snaps.</li>
                  <li xml:lang="es">Construye snaps.</li>
                  <li>Publish snaps to the store.</li>
                  <li xml:lang="es">Publica snaps en la tienda.</li>
                </ol>
              </description>
              <provides>
                <binary>snapcraft</binary>
              </provides>
            </component>
        """
        )

        with open(file_name, "w") as f:
            print(content, file=f)

        metadata = appstream.extract(file_name, workdir=".")

        self.assertThat(metadata.get_summary(), Equals("Create snaps"))
        self.assertThat(
            metadata.get_description(),
            Equals(
                textwrap.dedent(
                    """\
            Command Line Utility to create snaps.

            Features:

            1. Build snaps.
            2. Publish snaps to the store."""
                )
            ),
        )

    def test_appstream_with_ul_in_p(self):
        file_name = "snapcraft.appdata.xml"
        content = textwrap.dedent(
            """\
            <?xml version="1.0" encoding="UTF-8"?>
              <component type="desktop">
              <id>com.github.maoschanz.drawing</id>
              <metadata_license>CC0-1.0</metadata_license>
              <project_license>GPL-3.0-or-later</project_license>
              <content_rating type="oars-1.1"/>
              <name>Drawing</name>
              <name xml:lang="tr">Çizim</name>
              <name xml:lang="pt_BR">Drawing</name>
              <summary>A drawing application for the GNOME desktop</summary>
              <summary xml:lang="pt_BR">Uma aplicacao de desenho para o ambiente GNOME</summary>
              <summary xml:lang="nl">Een tekenprogramma voor de GNOME-werkomgeving</summary>
              <description>
                <p>"Drawing" is a basic image editor, supporting PNG, JPEG and BMP file types.</p>
                <p xml:lang="pt_BR">"Drawing" e um simples editor de imagens, que suporta arquivos PNG,JPEG e BMP</p>
                <p xml:lang="nl">"Tekenen" is een eenvoudige afbeeldingsbewerker, met ondersteuning voor PNG, JPEG en BMP.</p>
                <p xml:lang="fr">"Dessin" est un éditeur d'images basique, qui supporte les fichiers de type PNG, JPEG ou BMP.</p>
                <p>It allows you to draw or edit pictures with tools such as:
                  <ul>
                    <li>Pencil (with various options)</li>
                    <li xml:lang="pt_BR">Lápis (Com varias opções)</li>
                    <li xml:lang="nl">Potlood (verschillende soorten)</li>
                    <li xml:lang="fr">Crayon (avec diverses options)</li>
                    <li>Selection (cut/copy/paste/drag/…)</li>
                    <li xml:lang="tr">Seçim (kes/kopyala/yapıştır/sürükle /…)</li>
                    <li xml:lang="ru">Выделение (вырезать/копировать/вставить/перетащить/…)</li>
                    <li xml:lang="pt_BR">Seleção (cortar/copiar/colar/arrastar/…)</li>
                    <li xml:lang="nl">Selectie (knippen/kopiëren/plakken/verslepen/...)</li>
                    <li xml:lang="it">Selezione (taglia/copia/incolla/trascina/…)</li>
                    <li xml:lang="he">בחירה (חתיכה/העתקה/הדבקה/גרירה/...)</li>
                    <li xml:lang="fr">Sélection (copier/coller/déplacer/…)</li>
                    <li xml:lang="es">Selección (cortar/copiar/pegar/arrastrar/…)</li>
                    <li xml:lang="de_DE">Auswahl (Ausschneiden/Kopieren/Einfügen/Ziehen/...)</li>
                    <li>Line, Arc (with various options)</li>
                    <li xml:lang="pt_BR">Linha, Arco (com varias opcoes)</li>
                    <li xml:lang="nl">Lijn, Boog (verschillende soorten)</li>
                    <li xml:lang="fr">Trait, Arc (avec diverses options)</li>
                    <li>Shapes (rectangle, circle, polygon, …)</li>
                    <li xml:lang="pt_BR">Formas (retângulo, circulo, polígono, …)</li>
                    <li xml:lang="nl">Vormen (vierkant, cirkel, veelhoek, ...)</li>
                    <li xml:lang="fr">Formes (rectangle, cercle, polygone, …)</li>
                    <li>Text insertion</li>
                    <li xml:lang="pt_BR">Inserção de texto</li>
                    <li xml:lang="nl">Tekst invoeren</li>
                    <li xml:lang="fr">Insertion de texte</li>
                    <li>Resizing, cropping, rotating</li>
                    <li xml:lang="pt_BR">Redimencionar, cortar, rotacionar</li>
                    <li xml:lang="nl">Afmetingen wijzigen, bijsnijden, draaien</li>
                    <li xml:lang="fr">Redimensionnement, rognage, rotation</li>
                  </ul>
                </p>
              </description>
              </component>
        """
        )

        with open(file_name, "w") as f:
            print(content, file=f)

        metadata = appstream.extract(file_name, workdir=".")

        self.expectThat(
            metadata.get_summary(),
            Equals("A drawing application for the GNOME desktop"),
        )
        self.expectThat(
            metadata.get_description(),
            Equals(
                textwrap.dedent(
                    """\
                "Drawing" is a basic image editor, supporting PNG, JPEG and BMP file types.

                It allows you to draw or edit pictures with tools such as:

                - Pencil (with various options)
                - Selection (cut/copy/paste/drag/…)
                - Line, Arc (with various options)
                - Shapes (rectangle, circle, polygon, …)
                - Text insertion
                - Resizing, cropping, rotating"""
                )
            ),
        )

    def test_appstream_multilang_title(self):
        file_name = "foliate.appdata.xml"
        content = textwrap.dedent(
            """\
            <?xml version="1.0" encoding="UTF-8"?>
            <component type="desktop">
            <name>Foliate</name>
            <name xml:lang="id_ID">Foliate_id</name>
            <name xml:lang="pt_BR">Foliate_pt</name>
            <name xml:lang="ru_RU">Foliate_ru</name>
            <name xml:lang="nl_NL">Foliate_nl</name>
            <name xml:lang="fr_FR">Foliate_fr</name>
            <name xml:lang="cs_CS">Foliate_cs</name>
            </component>
        """
        )

        with open(file_name, "w") as f:
            print(content, file=f)

        metadata = appstream.extract(file_name, workdir=".")

        self.expectThat(metadata.get_title(), Equals("Foliate"))

    def test_appstream_release(self):
        file_name = "foliate.appdata.xml"
        content = textwrap.dedent(
            """\
            <?xml version="1.0" encoding="UTF-8"?>
            <component type="desktop">
            <releases>
                <release version="1.5.3" date="2019-07-25">
                <description>
                    <ul>
                    <li>Fixed Flatpak version not being able to open .mobi, .azw, and .azw3 files</li>
                    <li>Improved Wiktionary lookup, now with links and example sentences</li>
                    <li>Improved popover footnote extraction and formatting</li>
                    <li>Added option to export annotations to BibTeX</li>
                    </ul>
                </description>
                </release>
                <release version="1.5.2" date="2019-07-19">
                <description>
                    <ul>
                    <li>Fixed table of contents navigation not working with some books</li>
                    <li>Fixed not being able to zoom images with Kindle books</li>
                    <li>Fixed not being able to open books with .epub3 filename extension</li>
                    <li>Fixed temporary directory not being cleaned after closing</li>
                    </ul>
                </description>
                </release>
                <release version="1.5.1" date="2019-07-17">
                <description>
                    <ul>
                    <li>Fixed F9 shortcut not working</li>
                    <li>Updated translations</li>
                    </ul>
                </description>
                </release>
            </releases>
            </component>
        """
        )

        with open(file_name, "w") as f:
            print(content, file=f)

        metadata = appstream.extract(file_name, workdir=".")

        self.expectThat(metadata.get_version(), Equals("1.5.3"))

    def test_appstream_em(self):
        file_name = "foliate.appdata.xml"
        content = textwrap.dedent(
            """\
            <?xml version="1.0" encoding="UTF-8"?>
              <component type="desktop">
              <id>com.github.maoschanz.drawing</id>
              <metadata_license>CC0-1.0</metadata_license>
              <project_license>GPL-3.0-or-later</project_license>
              <content_rating type="oars-1.1"/>
              <name>Drawing</name>
              <description>
                <p>Command Line Utility to <em>create snaps</em> quickly.</p>
                <p xml:lang="es">Aplicativo de línea de comandos para crear snaps.</p>
                <p>Ordered Features:</p>
                <p xml:lang="es">Funciones:</p>
                <ol>
                  <li><em>Build snaps</em>.</li>
                  <li xml:lang="es">Construye snaps.</li>
                  <li>Publish snaps to the store.</li>
                  <li xml:lang="es">Publica snaps en la tienda.</li>
                </ol>
                <p>Unordered Features:</p>
                <ul>
                  <li><em>Build snaps</em>.</li>
                  <li xml:lang="es">Construye snaps.</li>
                  <li>Publish snaps to the store.</li>
                  <li xml:lang="es">Publica snaps en la tienda.</li>
                </ul>
              </description>
              </component>
        """
        )

        with open(file_name, "w") as f:
            print(content, file=f)

        metadata = appstream.extract(file_name, workdir=".")

        self.expectThat(
            metadata.get_description(),
            Equals(
                textwrap.dedent(
                    """\
            Command Line Utility to _create snaps_ quickly.

            Ordered Features:

            1. _Build snaps_.
            2. Publish snaps to the store.

            Unordered Features:

            - _Build snaps_.
            - Publish snaps to the store."""
                )
            ),
        )

    def test_appstream_code_tags_not_swallowed(self):
        file_name = "foliate.appdata.xml"
        content = textwrap.dedent(
            """\
            <?xml version="1.0" encoding="UTF-8"?>
              <component type="desktop">
              <id>com.github.maoschanz.drawing</id>
              <metadata_license>CC0-1.0</metadata_license>
              <project_license>GPL-3.0-or-later</project_license>
              <content_rating type="oars-1.1"/>
              <name>Drawing</name>
              <description>
                <p>Command Line Utility to <code>create snaps</code> quickly.</p>
                <p xml:lang="es">Aplicativo de línea de comandos para crear snaps.</p>
                <p>Ordered Features:</p>
                <p xml:lang="es">Funciones:</p>
                <ol>
                  <li><code>Build snaps</code>.</li>
                  <li xml:lang="es">Construye snaps.</li>
                  <li>Publish snaps to the store.</li>
                  <li xml:lang="es">Publica snaps en la tienda.</li>
                </ol>
                <p>Unordered Features:</p>
                <ul>
                  <li><code>Build snaps</code>.</li>
                  <li xml:lang="es">Construye snaps.</li>
                  <li>Publish snaps to the store.</li>
                  <li xml:lang="es">Publica snaps en la tienda.</li>
                </ul>
              </description>
              </component>
        """
        )

        with open(file_name, "w") as f:
            print(content, file=f)

        metadata = appstream.extract(file_name, workdir=".")

        self.expectThat(
            metadata.get_description(),
            Equals(
                textwrap.dedent(
                    """\
            Command Line Utility to create snaps quickly.

            Ordered Features:

            1. Build snaps.
            2. Publish snaps to the store.

            Unordered Features:

            - Build snaps.
            - Publish snaps to the store."""
                )
            ),
        )

    def test_appstream_with_comments(self):
        file_name = "foo.appdata.xml"
        content = textwrap.dedent(
            """\
            <?xml version="1.0" encoding="UTF-8"?>
              <component type="desktop">
              <id>com.github.maoschanz.drawing</id>
              <metadata_license>CC0-1.0</metadata_license>
              <project_license>GPL-3.0-or-later</project_license>
              <content_rating type="oars-1.1"/>
              <!-- TRANSLATORS: the application name -->
              <name>Drawing</name>
              <!-- TRANSLATORS: one-line description for the app -->
              <summary>Draw stuff</summary>
              <description>
                <!-- TRANSLATORS: AppData description marketing paragraph -->
                <p>Command Line Utility to create snaps quickly.</p>
                <p xml:lang="es">Aplicativo de línea de comandos para crear snaps.</p>
                <p>Ordered Features:</p>
                <p xml:lang="es">Funciones:</p>
                <ol>
                  <li>Build snaps.</li>
                  <li xml:lang="es">Construye snaps.</li>
                  <li>Publish snaps to the store.</li>
                  <li xml:lang="es">Publica snaps en la tienda.</li>
                </ol>
                <p>Unordered Features:</p>
                <ul>
                  <li>Build snaps.</li>
                  <li xml:lang="es">Construye snaps.</li>
                  <li>Publish snaps to the store.</li>
                  <li xml:lang="es">Publica snaps en la tienda.</li>
                </ul>
              </description>
              </component>
        """
        )

        with open(file_name, "w") as f:
            print(content, file=f)

        metadata = appstream.extract(file_name, workdir=".")

        self.expectThat(
            metadata.get_description(),
            Equals(
                textwrap.dedent(
                    """\
            Command Line Utility to create snaps quickly.

            Ordered Features:

            1. Build snaps.
            2. Publish snaps to the store.

            Unordered Features:

            - Build snaps.
            - Publish snaps to the store."""
                )
            ),
        )


class AppstreamUnhandledFileTestCase(unit.TestCase):
    def test_unhandled_file_test_case(self):
        raised = self.assertRaises(
            _errors.UnhandledFileError, appstream.extract, "unhandled-file", workdir="."
        )

        self.assertThat(raised.path, Equals("unhandled-file"))
        self.assertThat(raised.extractor_name, Equals("appstream"))


class TestAppstreamLaunchable:

    scenarios = (
        (
            "usr/share",
            {
                "desktop_file_path": "usr/share/applications/com.example.test/app.desktop"
            },
        ),
        (
            "usr/local/share",
            {
                "desktop_file_path": "usr/local/share/applications/com.example.test/app.desktop"
            },
        ),
    )

    def test(self, tmp_work_path, desktop_file_path):
        appstream_file = "foo.metainfo.xml"
        with open(appstream_file, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                <?xml version="1.0" encoding="UTF-8"?>
                <component>
                  <launchable type="desktop-id">
                    com.example.test-app.desktop
                  </launchable>
                </component>"""
                )
            )

        _create_desktop_file(desktop_file_path)

        extracted = appstream.extract("foo.metainfo.xml", workdir=".")

        assert extracted.get_desktop_file_paths() == [desktop_file_path]


class TestAppstreamLegacyDesktop:

    scenarios = (
        (
            "usr/share",
            {
                "desktop_file_path": "usr/share/applications/com.example.test/app.desktop"
            },
        ),
        (
            "usr/local/share",
            {
                "desktop_file_path": "usr/local/share/applications/com.example.test/app.desktop"
            },
        ),
    )

    def test_launchable(self, tmp_work_path, desktop_file_path):
        with open("foo.metainfo.xml", "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                <?xml version="1.0" encoding="UTF-8"?>
                <component type="desktop">
                  <id>com.example.test-app.desktop</id>
                </component>"""
                )
            )

        _create_desktop_file(desktop_file_path)

        extracted = appstream.extract("foo.metainfo.xml", workdir=".")

        assert extracted.get_desktop_file_paths() == [desktop_file_path]

    def test_appstream_no_desktop_suffix(self, tmp_work_path, desktop_file_path):
        with open("foo.metainfo.xml", "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                <?xml version="1.0" encoding="UTF-8"?>
                <component type="desktop">
                  <id>com.example.test-app</id>
                </component>"""
                )
            )

        _create_desktop_file(desktop_file_path)

        extracted = appstream.extract("foo.metainfo.xml", workdir=".")

        extracted.get_desktop_file_paths() == [desktop_file_path]


class AppstreamMultipleLaunchableTestCase(unit.TestCase):
    def test_appstream_with_multiple_launchables(self):
        with open("foo.metainfo.xml", "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                <?xml version="1.0" encoding="UTF-8"?>
                <component>
                  <launchable type="desktop-id">
                    com.example.test-app1.desktop
                  </launchable>
                  <launchable type="test-wrong-type">
                    dummy
                  </launchable>
                  <launchable type="desktop-id">
                    com.example.test-app2.desktop
                  </launchable>
                  <launchable type="desktop-id">
                    unexisting
                  </launchable>
                </component>"""
                )
            )

        _create_desktop_file(
            "usr/local/share/applications/com.example.test/app1.desktop"
        )
        _create_desktop_file(
            "usr/local/share/applications/com.example.test/app2.desktop"
        )

        expected = [
            "usr/local/share/applications/com.example.test/app1.desktop",
            "usr/local/share/applications/com.example.test/app2.desktop",
        ]
        extracted = appstream.extract("foo.metainfo.xml", workdir=".")

        self.assertThat(extracted.get_desktop_file_paths(), Equals(expected))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys as _sys

if _sys.platform == "linux":
    from snapcraft.internal.sources import Bazaar  # noqa
    from snapcraft.internal.sources import Deb  # noqa
    from snapcraft.internal.sources import Git  # noqa
    from snapcraft.internal.sources import Local  # noqa
    from snapcraft.internal.sources import Mercurial  # noqa
    from snapcraft.internal.sources import Rpm  # noqa
    from snapcraft.internal.sources import Script  # noqa
    from snapcraft.internal.sources import Subversion  # noqa
    from snapcraft.internal.sources import Tar  # noqa
    from snapcraft.internal.sources import Zip  # noqa
    from snapcraft.internal.sources import get  # noqa



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
from textwrap import dedent
from unittest import mock

import fixtures
from simplejson.scanner import JSONDecodeError
from testtools.matchers import Contains, Equals

from snapcraft import storeapi

from . import FakeStoreCommandsBaseTestCase, get_sample_key


class RegisterKeyTestCase(FakeStoreCommandsBaseTestCase):
    def test_register_key(self):
        result = self.run_command(
            ["register-key", "default"], input="user@example.com\nsecret\n"
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains("Registering key ..."))
        self.assertThat(
            result.output,
            Contains(
                'Done. The key "default" ({}) may be used to sign your '
                "assertions.".format(get_sample_key("default")["sha3-384"])
            ),
        )
        self.fake_store_login.mock.assert_called_once_with(
            email="user@example.com",
            password="secret",
            acls=["modify_account_key"],
            packages=None,
            channels=None,
            expires=None,
            save=False,
            config_fd=None,
        )
        self.fake_store_register_key.mock.call_once_with(
            dedent(
                """\
                type: account-key-request
                account-id: abcd
                name: default
                public-key-sha3-384: {}
                """
            ).format(get_sample_key("default")["sha3-384"])
        )

    def test_register_key_no_keys(self):
        self.fake_check_output.mock.side_effect = [json.dumps([]).encode()]

        raised = self.assertRaises(
            storeapi.errors.NoKeysError, self.run_command, ["register-key"]
        )

        self.assertThat(str(raised), Contains("You have no usable keys"))

    def test_register_key_no_keys_with_name(self):
        raised = self.assertRaises(
            storeapi.errors.NoSuchKeyError,
            self.run_command,
            ["register-key", "nonexistent"],
        )

        self.assertThat(
            str(raised), Contains("You have no usable key named 'nonexistent'")
        )

    def test_register_key_login_failed(self):
        self.fake_store_login.mock.side_effect = storeapi.http_clients.errors.InvalidCredentialsError(
            "error"
        )

        raised = self.assertRaises(
            storeapi.http_clients.errors.InvalidCredentialsError,
            self.run_command,
            ["register-key", "default"],
            input="user@example.com\nsecret\n",
        )

        assert (
            str(raised) == 'Invalid credentials: error. Have you run "snapcraft login"?'
        )

    def test_register_key_account_info_failed(self):
        response = mock.Mock()
        response.json.side_effect = JSONDecodeError("mock-fail", "doc", 1)
        response.status_code = 500
        response.reason = "Internal Server Error"
        self.fake_store_account_info.mock.side_effect = storeapi.errors.StoreAccountInformationError(
            response
        )

        # Fake the login check
        self.useFixture(fixtures.MockPatch("snapcraft._store.login", return_value=True))

        raised = self.assertRaises(
            storeapi.errors.StoreAccountInformationError,
            self.run_command,
            ["register-key", "default"],
        )

        self.assertThat(
            str(raised),
            Equals(
                "Error fetching account information from store: "
                "500 Internal Server Error"
            ),
        )

    def test_register_key_failed(self):
        response = mock.Mock()
        response.json.side_effect = JSONDecodeError("mock-fail", "doc", 1)
        response.status_code = 500
        response.reason = "Internal Server Error"
        self.fake_store_register_key.mock.side_effect = storeapi.errors.StoreKeyRegistrationError(
            response
        )

        raised = self.assertRaises(
            storeapi.errors.StoreKeyRegistrationError,
            self.run_command,
            ["register-key", "default"],
            input="user@example.com\nsecret\n",
        )

        self.assertThat(
            str(raised), Equals("Key registration failed: 500 Internal Server Error")
        )

    def test_register_key_select_key(self):
        result = self.run_command(
            ["register-key"], input="x\n2\nuser@example.com\nsecret\n"
        )

        self.assertThat(
            result.output,
            Contains(
                dedent(
                    """\
            Select a key:

              Number  Name     SHA3-384 fingerprint
                   1  default  {default_sha3_384}
                   2  another  {another_sha3_384}
            """
                ).format(
                    default_sha3_384=get_sample_key("default")["sha3-384"],
                    another_sha3_384=get_sample_key("another")["sha3-384"],
                )
            ),
        )
        self.assertThat(
            self.fake_logger.output,
            Contains(
                'Done. The key "another" ({}) may be used to sign your '
                "assertions.\n".format(get_sample_key("another")["sha3-384"])
            ),
        )

        self.fake_store_register_key.mock.assert_called_once_with(
            dedent(
                """\
            type: account-key-request
            account-id: abcd
            name: another
            public-key-sha3-384: {}
            """
            ).format(get_sample_key("another")["sha3-384"])
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import fileinput
import os
import subprocess

import testscenarios
from testtools.matchers import Contains, Equals

from tests import fixture_setup, integration
from tests.integration import repo


def _construct_scenarios():
    main_scenarios = {
        "build-snap-grammar": True,
        "build-snap-grammar-try": True,
        "build-snap-grammar-try-skip": False,
        "build-snap-grammar-try-else": True,
        "build-snap-grammar-on": False,
        "build-snap-grammar-on-else": True,
    }

    # Just some combinations
    channel_scenarios = ["", "/stable", "/latest/stable"]

    all_scenarios = []
    for project, expected_install in main_scenarios.items():
        for channel in channel_scenarios:
            d = dict(project=project, hello_installed=expected_install, channel=channel)
            scenario = ("{}{}".format(project, channel), d)
            all_scenarios.append(scenario)

    return all_scenarios


class BuildSnapGrammarTestCase(testscenarios.WithScenarios, integration.TestCase):

    scenarios = _construct_scenarios()

    def setUp(self):
        super().setUp()
        # We cannot install snaps on the adt test bed at this time.
        # - Mount snap "core" (2775) ([start snap-core-2775.mount] \
        # failed with exit status 1: Job for snap-core-2775.mount failed.
        if os.environ.get("ADT_TEST") and self.deb_arch == "armhf":
            self.skipTest("snap installation not working well with adt on armhf")
        self.useFixture(fixture_setup.WithoutSnapInstalled("hello"))

    def _add_channel_information_to_hello(self):
        replacement = "- hello{}".format(self.channel)
        with fileinput.input("snapcraft.yaml", inplace=True) as input_file:
            for line in input_file:
                print(line.replace("- hello", replacement), end="")

    def test_grammar(self):
        self.copy_project_to_cwd(self.project)
        self._add_channel_information_to_hello()

        self.run_snapcraft("pull")

        self.assertThat(repo.is_snap_installed("hello"), Equals(self.hello_installed))


class BuildSnapGrammarErrorsTestCase(integration.TestCase):
    def test_on_other_arch_else_fail(self):
        """Test that 'on' fails with an error if it hits an 'else fail'."""

        exception = self.assertRaises(
            subprocess.CalledProcessError,
            self.run_snapcraft,
            ["pull"],
            "build-snap-grammar-fail",
        )

        self.assertThat(
            exception.output,
            Contains("Unable to satisfy 'on other-arch', failure forced"),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from textwrap import dedent

import fixtures
from testtools.matchers import Equals, HasLength

from snapcraft.internal import errors, mountinfo
from tests import unit


class MountInfoTestCase(unit.TestCase):
    def _write_mountinfo(self, contents):
        path = "mountinfo"
        with open(path, "w") as f:
            f.write(contents)
        return path

    def test_mountinfo_by_root(self):
        mounts = mountinfo.MountInfo(
            mountinfo_file=self._write_mountinfo(
                dedent(
                    """\
                23 28 0:4 / /proc rw,nosuid,nodev,noexec,relatime shared:14 - proc proc rw
                1341 28 7:6 / /snap/snapcraft/1 ro,nodev,relatime shared:39 - squashfs /dev/loop6 ro
                1455 28 253:0 /test-snap/prime /snap/test-snap/x1 ro,relatime shared:1 - ext4 /dev/mapper/foo rw,errors=remount-ro,data=ordered
                """
                )
            )
        )  # noqa

        root_mounts = mounts.for_root("/")
        for mount_point in ("/proc", "/snap/snapcraft/1"):
            self.assertTrue(
                any(m for m in root_mounts if m.mount_point == mount_point),
                "Expected {!r} to be included in root mounts".format(mount_point),
            )

        test_snap_mounts = mounts.for_root("/test-snap/prime")
        self.assertThat(test_snap_mounts, HasLength(1))
        self.expectThat(test_snap_mounts[0].mount_point, Equals("/snap/test-snap/x1"))

    def test_mountinfo_by_mount_point(self):
        mounts = mountinfo.MountInfo(
            mountinfo_file=self._write_mountinfo(
                dedent(
                    """\
                23 28 0:4 / /proc rw,nosuid,nodev,noexec,relatime shared:14 - proc proc rw
                1341 28 7:6 / /snap/snapcraft/1 ro,nodev,relatime shared:39 - squashfs /dev/loop6 ro
                1455 28 253:0 /test-snap/prime /snap/test-snap/x1 ro,relatime shared:1 - ext4 /dev/mapper/foo rw,errors=remount-ro,data=ordered
                """
                )
            )
        )  # noqa

        mount = mounts.for_mount_point("/proc")
        self.assertThat(mount.mount_id, Equals("23"))
        self.assertThat(mount.parent_id, Equals("28"))
        self.assertThat(mount.st_dev, Equals("0:4"))
        self.assertThat(mount.root, Equals("/"))
        self.assertThat(mount.mount_point, Equals("/proc"))
        self.assertThat(mount.mount_options, Equals("rw,nosuid,nodev,noexec,relatime"))
        self.assertThat(mount.optional_fields, Equals(["shared:14"]))
        self.assertThat(mount.filesystem_type, Equals("proc"))
        self.assertThat(mount.mount_source, Equals("proc"))
        self.assertThat(mount.super_options, Equals("rw"))

        mount = mounts.for_mount_point("/snap/snapcraft/1")
        self.assertThat(mount.mount_id, Equals("1341"))
        self.assertThat(mount.parent_id, Equals("28"))
        self.assertThat(mount.st_dev, Equals("7:6"))
        self.assertThat(mount.root, Equals("/"))
        self.assertThat(mount.mount_point, Equals("/snap/snapcraft/1"))
        self.assertThat(mount.mount_options, Equals("ro,nodev,relatime"))
        self.assertThat(mount.optional_fields, Equals(["shared:39"]))
        self.assertThat(mount.filesystem_type, Equals("squashfs"))
        self.assertThat(mount.mount_source, Equals("/dev/loop6"))
        self.assertThat(mount.super_options, Equals("ro"))

        mount = mounts.for_mount_point("/snap/test-snap/x1")
        self.assertThat(mount.mount_id, Equals("1455"))
        self.assertThat(mount.parent_id, Equals("28"))
        self.assertThat(mount.st_dev, Equals("253:0"))
        self.assertThat(mount.root, Equals("/test-snap/prime"))
        self.assertThat(mount.mount_point, Equals("/snap/test-snap/x1"))
        self.assertThat(mount.mount_options, Equals("ro,relatime"))
        self.assertThat(mount.optional_fields, Equals(["shared:1"]))
        self.assertThat(mount.filesystem_type, Equals("ext4"))
        self.assertThat(mount.mount_source, Equals("/dev/mapper/foo"))
        self.assertThat(
            mount.super_options, Equals("rw,errors=remount-ro,data=ordered")
        )

    def test_mountinfo_missing_root(self):
        mounts = mountinfo.MountInfo(mountinfo_file=self._write_mountinfo(""))
        raised = self.assertRaises(
            errors.RootNotMountedError, mounts.for_root, "test-root"
        )
        self.assertThat(raised.root, Equals("test-root"))

    def test_mountinfo_missing_mount_point(self):
        mounts = mountinfo.MountInfo(mountinfo_file=self._write_mountinfo(""))
        raised = self.assertRaises(
            errors.MountPointNotFoundError, mounts.for_mount_point, "test-root"
        )
        self.assertThat(raised.mount_point, Equals("test-root"))

    def test_invalid_mountinfo(self):
        self.fake_logger = fixtures.FakeLogger(level=logging.WARN)
        self.useFixture(self.fake_logger)

        mountinfo.MountInfo(mountinfo_file=self._write_mountinfo(dedent("I'm invalid")))

        # Assert that a warning was logged
        self.assertThat(
            self.fake_logger.output,
            Equals("Unable to parse mountinfo row: I'm invalid\n"),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from typing import Dict
from urllib.parse import urljoin

import requests

from . import constants, errors
from ._requests import Requests
from .info import SnapInfo


logger = logging.getLogger(__name__)


class SnapAPI(Requests):
    """The Snap API is used to query snaps.

    This is an interface to query that API which is documented
    at http://api.snapcraft.io/docs/.
    """

    def __init__(self, client):
        self._client = client
        self._root_url = os.environ.get("STORE_API_URL", constants.STORE_API_URL)

    def _request(self, method, urlpath, **kwargs) -> requests.Response:
        url = urljoin(self._root_url, urlpath)
        return self._client.request(method, url, **kwargs)

    def _get_default_headers(self, api="v2"):
        """Return default headers for CPI requests.

        Tries to build an 'Authorization' header with local credentials
        if they are available.
        Also pin specific branded store if `SNAPCRAFT_UBUNTU_STORE`
        environment is set.
        """
        headers = {}

        brand_store = os.getenv("SNAPCRAFT_UBUNTU_STORE")
        if brand_store:
            if api == "v2":
                headers["Snap-Device-Store"] = brand_store
            elif api == "v1":
                headers["X-Ubuntu-Store"] = brand_store
            else:
                logger.warning("Incorrect API version passed: {!r}.".format(api))

        return headers

    def get_info(self, snap_name: str, *, arch: str = None) -> SnapInfo:
        """Get the information for the specified snap.

        :param str snap_name: Name of the snap.
        :param str arch: Architecture of the snap (none by default).

        :returns: information for the snap.
        :rtype: SnapInfo
        """
        headers = self._get_default_headers()
        headers.update(
            {
                "Accept": "application/json",
                "Snap-Device-Series": constants.DEFAULT_SERIES,
            }
        )

        params = dict()
        params[
            "fields"
        ] = "channel-map,snap-id,name,publisher,confinement,revision,download"
        if arch is not None:
            params["architecture"] = arch
        logger.debug("Getting information for {}".format(snap_name))
        url = "/v2/snaps/info/{}".format(snap_name)
        resp = self.get(url, headers=headers, params=params)

        if resp.status_code == 404:
            raise errors.SnapNotFoundError(snap_name=snap_name, arch=arch)
        resp.raise_for_status()

        return SnapInfo(resp.json())

    def get_assertion(
        self, assertion_type: str, snap_id: str
    ) -> Dict[str, Dict[str, str]]:
        """Get the assertion of assertion_type for the specified snap_id.

        :param str assertion_type: The type of assertion.
        :param str snap_id: The ID of the snap.

        :returns: Assertion for the snap.
        """
        headers = self._get_default_headers(api="v1")
        logger.debug("Getting snap-declaration for {}".format(snap_id))
        url = f"/api/v1/snaps/assertions/{assertion_type}/{constants.DEFAULT_SERIES}/{snap_id}"
        response = self.get(url, headers=headers)
        if response.status_code != 200:
            raise errors.SnapNotFoundError(snap_id=snap_id)
        return response.json()
